from uuid import uuid4

from amino import Dat, Maybe, Just, Either, Nothing
from amino.case import Case

from chiasma.data.view_tree import ViewTree, LayoutNode, PaneNode, SubUiNode
from chiasma.util.id import IdentSpec, ensure_ident_or_generate

from myo.util import Ident
from myo.ui.data.view import Pane, Layout


class Window(Dat['Window']):

    @staticmethod
    def cons(
            ident: IdentSpec=None,
            layout: LayoutNode[Layout, Pane]=None,
    ) -> 'Window':
        return Window(
            ensure_ident_or_generate(ident),
            layout or ViewTree.layout(),
        )

    def __init__(self, ident: Ident, layout: LayoutNode[Layout, Pane]) -> None:
        self.ident = ident
        self.layout = layout


class FindPrincipal(Case, alg=ViewTree):

    def layout_node(self, node: LayoutNode) -> Maybe[Pane]:
        return node.sub.find_map(self)

    def pane_node(self, node: PaneNode) -> Maybe[Pane]:
        return Just(node.data)

    def sub_ui_node(self, node: SubUiNode) -> Maybe[Pane]:
        return Nothing


def find_principal(window: Window) -> Either[str, Pane]:
    return FindPrincipal.match(window.layout).to_either(f'window contains no pane')


__all__ = ('Window', 'find_principal')



from typing import TypeVar

from amino import do, Do, Maybe, List, Either, __, Boolean, Nil
from amino.dat import Dat
from amino.lenses.lens import lens
from amino.boolean import false

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome import Ribosome
from ribosome.compute.ribosome_api import Ribo

from myo.util import Ident
from myo.data.command import Command, SystemInterpreter, VimInterpreter, ShellInterpreter
from myo.env import Env
from myo.config.component import MyoComponent
from myo.components.command.data import CommandData

A = TypeVar('A')
CommandRiboState = NS[Ribosome[Env, MyoComponent, CommandData], A]


class AddVimCommandOptions(Dat['AddVimCommandOptions']):

    def __init__(
            self,
            ident: Maybe[Ident],
            line: Maybe[str],
            lines: Maybe[List[str]],
            silent: Maybe[Boolean],
            target: Maybe[Ident],
            langs: Maybe[List[str]],
    ) -> None:
        self.ident = ident
        self.line = line
        self.lines = lines
        self.silent = silent
        self.target = target
        self.langs = langs


class AddSystemCommandOptions(Dat['AddSystemCommandOptions']):

    def __init__(
            self,
            ident: Maybe[Ident],
            line: Maybe[str],
            lines: Maybe[List[str]],
            target: Maybe[Ident],
            langs: Maybe[List[str]],
    ) -> None:
        self.ident = ident
        self.line = line
        self.lines = lines
        self.target = target
        self.langs = langs


class AddShellCommandOptions(Dat['AddShellCommandOptions']):

    def __init__(
            self,
            ident: Maybe[Ident],
            line: Maybe[str],
            lines: Maybe[List[str]],
            target: Ident,
            langs: Maybe[List[str]],
    ) -> None:
        self.ident = ident
        self.line = line
        self.lines = lines
        self.target = target
        self.langs = langs


def cons_vim_command(options: AddVimCommandOptions) -> Either[str, Command]:
    return Command.cons(
        options.ident | Ident.generate,
        VimInterpreter(options.silent | false, options.target),
        options.lines.o(options.line / List) | Nil,
        options.langs | List('vim'),
    )


def cons_system_command(options: AddSystemCommandOptions) -> Either[str, Command]:
    return Command.cons(
        options.ident | Ident.generate,
        SystemInterpreter(options.target),
        options.lines.o(options.line / List) | Nil,
        options.langs | Nil,
    )


def cons_shell_command(options: AddShellCommandOptions) -> Command:
    return Command.cons(
        options.ident | Ident.generate,
        ShellInterpreter(options.target),
        options.lines.o(options.line / List) | Nil,
        options.langs | Nil,
    )


@prog.unit
@do(NS[Ribosome[Env, MyoComponent, CommandData], None])
def add_vim_command(options: AddVimCommandOptions) -> Do:
    cmd = cons_vim_command(options)
    yield Ribo.modify_comp(lens.commands.modify(__.cat(cmd)))


@prog.unit
@do(NS[Ribosome[Env, MyoComponent, CommandData], None])
def add_system_command(options: AddSystemCommandOptions) -> Do:
    cmd = cons_system_command(options)
    yield Ribo.modify_comp(lens.commands.modify(__.cat(cmd)))


@prog.unit
@do(NS[Ribosome[Env, MyoComponent, CommandData], None])
def add_shell_command(options: AddShellCommandOptions) -> Do:
    cmd = cons_shell_command(options)
    yield Ribo.modify_comp(lens.commands.modify(__.cat(cmd)))


__all__ = ('add_system_command', 'add_vim_command')



from amino import do, Do
from amino.lenses.lens import lens

from ribosome.compute.api import prog

from chiasma.util.id import Ident
from chiasma.pane import pane_by_ident
from chiasma.io.state import TS
from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo

from myo.components.tmux.tpe import TmuxRibosome
from myo.tmux.pid import process_pid
from myo.util.process import kill_process
from myo.components.tmux.pane import pane_id_by_ident


@prog
@do(NS[TmuxRibosome, None])
def tmux_kill_pane(ident: Ident) -> Do:
    pane_id = yield pane_id_by_ident(ident)
    pane = yield Ribo.zoom_comp(pane_by_ident(ident).nvim.zoom(lens.views))
    pane_id = yield NS.m(pane.id, lambda: f'pane `{ident}` has no id')
    pid_m = yield TS.lift(process_pid(pane_id)).nvim
    pid = yield NS.m(pid_m, lambda: f'pane `{ident}` has no running process')
    # signal!
    yield NS.from_io(kill_process(pid))
    yield NS.unit


__all__ = ('tmux_kill_pane',)



from amino import do, List, Do

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.function import define_function


@do(NvimIO[None])
def mock_test_functions() -> Do:
    yield define_function('MyoTestDetermineRunner', List('file'), f'return "python"')
    yield define_function('MyoTestExecutable', List('runner'), f'return "python"')
    yield define_function('MyoTestBuildPosition', List('runner', 'pos'), f'return ["-c", "raise Exception(1)"]')
    yield define_function('MyoTestBuildArgs', List('runner', 'args'), f'return a:args')


__all__ = ('mock_test_functions',)



from typing import Callable

from myo.components.tmux.data import TmuxData
from chiasma.data.session import Session
from chiasma.commands.session import session_id
from chiasma.data.window import Window
from chiasma.commands.window import window_id
from chiasma.data.pane import Pane
from chiasma.commands.pane import pane_id

from amino import do, Do, List, Maybe

from ribosome.nvim.io.state import NS
from ribosome.config.component import ComponentData, comp_data
from ribosome.compute.api import prog

from myo.env import Env
from myo.data.info import InfoWidget


def format_id(id: Maybe[int], formatter: Callable[[int], str]) -> str:
    return id.map(lambda a: f' {formatter(a)}') | ''


def format_pane(pane: Pane) -> List[str]:
    return List(f'◳ {pane.ident.str}{format_id(pane.id, pane_id)}')


def format_window(window: Window) -> List[str]:
    return List(f'□ {window.ident.str}{format_id(window.id, window_id)}')


def format_session(session: Session) -> List[str]:
    return List(f'⬚ {session.ident.str}{format_id(session.id, session_id)}')


@prog
@do(NS[ComponentData[Env, TmuxData], InfoWidget])
def tmux_info() -> Do:
    s = yield comp_data()
    views = s.views
    sessions = (views.sessions // format_session).indent(1)
    windows = (views.windows // format_window).indent(1)
    panes = (views.panes // format_pane).indent(1)
    yield NS.pure(InfoWidget.cons(List('Tmux:') + sessions + windows + panes))

__all__ = ('tmux_info',)



from amino import Right, Nil
from amino.boolean import true

from chiasma.ui.view_geometry import ViewGeometry

from ribosome.config.setting import (float_setting, str_setting, int_setting, bool_setting, str_list_setting,
                                     json_setting)

state_dir_help = f'''
Stored data consists of:
* command history
* last test command line
* last shell used for testing
* layout state
'''
tmux_watcher_interval_help = f'''Tmux panes can be observed for events, like processes terminating.
This specifies the polling interval.
'''
tmux_socket_help = '''The tmux server can be chosen by its socket path. This is mainly intended for tests.
'''
vim_tmux_pane_help = '''Skip discovery of the tmux pane hosting neovim by its process id and use the supplied pane id
instead. Particularly helpful for tests.
'''
display_parse_result_help = '''After parsing the output of an executed command, display errors in a scratch buffer.
'''
auto_jump_help = '''When loading the output buffer or cycling through output events, jump to the code location without
hitting the jump key.
'''
vim_test_filename_modifier_help = '''A vim-test setting that is applied to file names.
'''
init_default_ui_help = '''Create space, window, layout and pane for vim and the default execution target at startup.
The default target pane will be positioned to the right of vim.
'''
test_ui_help = '''Tests run with vim-test will be executed in this ui. Can be `tmux` or `internal`.
'''
test_pane_help = '''Tests run with vim-test will be executed in this pane.
Defaults to `make`.
'''
test_shell_help = '''Tests run with vim-test will be executed in this shell.
Takes precedence over `test_pane`.
'''
test_langs_help = '''When parsing the output of a command executed with `MyoVimTest`, these languages are expected.
'''
load_history_help = '''The history of executed commands is persisted to disk. This setting controls whether it is
restored on startup.
'''
builtin_output_config_help = '''Command output parsing can be configured from multiple sources. If this setting is
`true`, myo's built-in configuration is used in addition to command-specific config.
'''
vim_pane_geometry_help = '''The vim pane is autodetected and -created, so its size, weight and position can be
overridden here. The format is:
```{
  "min_size": 50,
  "max_size": 50,
  "fixed_size": 50,
  "minimized_size": 50,
  "weight": 0.5,
  "position": 50,
}
```
Since the vim pane is wrapped in a vertical layout, these settings are actually applied to that, not the pane itself.
All settings are optional; the default is a weight of 0.5.
'''


tmux_watcher_interval = float_setting('tmux_watcher_interval', 'tmux process polling interval',
                                      tmux_watcher_interval_help, True, Right(1.0))
tmux_socket = str_setting('tmux_socket', 'tmux socket path', tmux_socket_help, True)
vim_tmux_pane = int_setting('vim_tmux_pane', 'vim tmux pane id', vim_tmux_pane_help, True)
display_parse_result = bool_setting('display_parse_result', 'display parse result', display_parse_result_help, True,
                                    Right(true))
auto_jump = bool_setting('auto_jump', 'jump when changing output events', auto_jump_help, True, Right(true))
vim_test_filename_modifier = str_setting('test#filename_modifier', 'vim-test filename modifier',
                                         vim_test_filename_modifier_help, False, Right(':.'))
init_default_ui = bool_setting('init_default_ui', 'initialize vim and make panes', init_default_ui_help, True,
                               Right(true))
test_ui = str_setting('test_ui', 'ui for running tests', test_ui_help, True, Right('tmux'))
test_pane = str_setting('test_pane', 'pane for running tests', test_pane_help, True, Right('make'))
test_shell = str_setting('test_shell', 'shell for running tests', test_shell_help, True)
test_langs = str_list_setting('test_langs', 'parsing langs for vim-test output', test_langs_help, True, Right(Nil))
load_history = bool_setting('load_history', 'load command history', load_history_help, True, Right(True))
builtin_output_config = bool_setting('builtin_output_config', 'use default output configs', builtin_output_config_help,
                                     True, Right(True))
vim_pane_geometry = json_setting(ViewGeometry)('vim_pane_geometry', 'geometry overrides for the vim pane',
                                               vim_pane_geometry_help, True, Right(ViewGeometry.cons(weight=0.5)))


__all__ = ('tmux_watcher_interval', 'tmux_socket', 'vim_tmux_pane', 'display_parse_result', 'auto_jump',
           'vim_test_filename_modifier', 'init_default_ui', 'test_ui', 'test_pane', 'test_langs', 'load_history',
           'builtin_output_config', 'test_shell', 'vim_pane_geometry',)



from amino import do, Do, _, List
from amino.state import State
from amino.logging import module_log

from ribosome.compute.api import prog
from ribosome.compute.prog import Prog
from ribosome.compute.program import Program, bind_nullary_program

from myo.config.handler import find_handlers
from myo.config.plugin_state import MyoState

log = module_log()


@prog
@do(State[MyoState, List[Program[None]]])
def init_handlers() -> Do:
    yield find_handlers(_.init)


@prog.do(None)
def init() -> Do:
    handlers = yield find_handlers(_.init)
    yield handlers.traverse(bind_nullary_program, Prog)
    yield Prog.unit


__all__ = ('init',)



from amino import do, Do, Map
from amino.logging import module_log

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo
from ribosome.util.menu.data import MenuLine, MenuQuitWith, MenuUnit, Menu, MenuConfig
from ribosome.util.menu.auto.run import auto_menu, selected_menu_lines
from ribosome.util.menu.auto.data import AutoS, AutoState
from ribosome.util.menu.run import run_menu_prog
from ribosome.compute.prog import Prog

from myo.components.command.compute.tpe import CommandRibosome
from myo.components.command.compute.run import run_command
from myo.data.command import HistoryEntry
from myo.components.command.data import CommandData

log = module_log()


@do(AutoS)
def history_menu_selected() -> Do:
    items = yield selected_menu_lines()
    return items.head.cata_f(
        lambda a: MenuQuitWith(run_command(a.meta.cmd)),
        lambda: MenuUnit(),
    )


@do(NS[CommandRibosome, Menu[AutoState[None, HistoryEntry, None], HistoryEntry, None]])
def setup_history_menu() -> Do:
    history = yield Ribo.inspect_comp(lambda a: a.history)
    lines = history.map(
        lambda a: MenuLine.cons(f'[{a.cmd.ident.str}] {a.cmd.lines.head.get_or_strict("")}', a)
    )
    return auto_menu(
        None,
        lines,
        MenuConfig.cons('myo command history', insert=False),
        Map({'<cr>': history_menu_selected}),
    )


@prog.do(None)
def history_menu() -> Do:
    menu = yield Ribo.lift(setup_history_menu(), CommandData)
    yield run_menu_prog(menu)
    yield Prog.unit


__all__ = ('history_menu',)



from amino import List

from ribosome.nvim.io.state import NS
from ribosome.nvim.syntax.syntax import Syntax
from ribosome.nvim.syntax.expr import SyntaxMatch, HiLink, SyntaxLiteral, SyntaxRegion, Highlight

from myo.components.command.compute.tpe import CommandRibosome
from myo.output.lang.scala.report import col_marker

error_end = f'\ze.*\(\|{col_marker}\)'
foundreqHead = 'type mismatch'
noInstanceMarker = '\s*!instance:'
notInScopeMarker = 'Variable not in scope:'
haskell_include = SyntaxLiteral('syntax include @haskell syntax/haskell.vim')
location = SyntaxMatch.cons('MyoLocation', '^.*.*$', 'skipwhite', 'skipnl',
                            contains='MyoPath,MyoLineNumber', nextgroup='MyoHsError')
path = SyntaxMatch.cons('MyoPath', '^.*\ze\( .*$\)\@=', 'contained', containedin='MyoLocation')
line_number = SyntaxMatch.cons('MyoLineNumber', '\( \)\@<=\zs\d\+\ze', 'contained', containedin='MyoLocation')
# function = SyntaxMatch.cons('MyoFunction', '\( \d\+ \)\@<=\zs.*$', 'contained', containedin='MyoLocation')
error = SyntaxRegion.cons(
    'MyoHsError',
    '^',
    error_end,
    'contained',
    'skipwhite',
    'skipnl',
    contains='MyoHsFoundReq,MyoHsNoInstance,MyoHsNotInScope',
)
foundreq = SyntaxRegion.cons(
    'MyoHsFoundReq',
    foundreqHead,
    error_end,
    'contained',
    'skipwhite',
    'skipnl',
    contains='MyoHsFound',
)
found = SyntaxMatch.cons('MyoHsFound', '^.*$', 'contained', 'skipnl', nextgroup='MyoHsReq')
req = SyntaxMatch.cons('MyoHsReq', '^.*$', 'contained')
code = SyntaxMatch.cons('MyoHsCode', '.*', 'contained', contains='@haskell')
noInstance = SyntaxRegion.cons(
    'MyoHsNoInstance',
    noInstanceMarker,
    error_end,
    'contained',
    contains='MyoHsNoInstanceHead',
)
noInstanceHead = SyntaxMatch.cons(
    'MyoHsNoInstanceHead',
    f'\s*{noInstanceMarker} .*$',
    'contained',
    'skipnl',
    contains='MyoHsNoInstanceBang',
    nextgroup='MyoHsNoInstanceDesc',
)
noInstanceBang = SyntaxMatch.cons('MyoHsNoInstanceBang', '!', 'contained', nextgroup='MyoHsNoInstanceKw')
noInstanceKw = SyntaxMatch.cons(
    'MyoHsNoInstanceKw',
    'instance\ze:',
    'contained',
    'skipwhite',
    nextgroup='MyoHsNoInstanceTrigger',
)
noInstanceTrigger = SyntaxMatch.cons('MyoHsNoInstanceTrigger', '.*', 'contained')
noInstanceDesc = SyntaxMatch.cons('MyoHsNoInstanceDesc', '.*', 'contained', 'skipnl', nextgroup='MyoLocation')
notInScope = SyntaxRegion.cons(
    'MyoHsNotInScope',
    notInScopeMarker,
    error_end,
    'contained',
    contains='MyoHsNotInScopeHead',
)
notInScopeHead = SyntaxMatch.cons(
    'MyoHsNotInScopeHead',
    f'\s*{notInScopeMarker}',
    'skipnl',
    'contained',
    nextgroup='MyoHsCode',
)

hl_path = HiLink('MyoPath', 'Directory')
hl_line_number = HiLink('MyoLineNumber', 'Directory')
hl_foundreq = HiLink('MyoHsFoundReq', 'Title')
hl_found = HiLink('MyoHsFound', 'Error')
hl_bang = HiLink('MyoHsNoInstanceBang', 'Error')
hl_kw = HiLink('MyoHsNoInstanceKw', 'Directory')
hl_notinscope = HiLink('MyoHsNotInScopeHead', 'Error')

hi_req = Highlight.cons('MyoHsReq', ctermfg=2, guifg='#719e07')
hi_trigger = Highlight.cons('MyoHsNoInstanceTrigger', ctermfg=3)

syntax = Syntax(
    List(haskell_include, location, path, line_number, error, foundreq, found, req, code, noInstance, noInstanceHead,
         noInstanceBang, noInstanceKw, noInstanceTrigger, noInstanceDesc, notInScope, notInScopeHead),
    List(hi_req, hi_trigger),
    List(hl_path, hl_line_number, hl_foundreq, hl_found, hl_bang, hl_kw, hl_notinscope),
)


def haskell_syntax() -> NS[CommandRibosome, Syntax]:
    return NS.pure(syntax)


__all__ = ('haskell_syntax',)



from amino import List, Dat, Maybe, Just, Nothing

from ribosome.config.component import Component
from ribosome.compute.program import Program

from myo.components.vim.compute.run import run, vim_can_run
from myo.config.component import MyoComponent
from myo.command.run_task import RunTask


def run_handler_for(task: RunTask) -> Maybe[Program]:
    return Just(run) if vim_can_run(task) else Nothing


class VimData(Dat['VimData']):

    @staticmethod
    def cons(
    ) -> 'VimData':
        return VimData(
        )

    def __init__(self) -> None:
        pass


vim = Component.cons(
    'vim',
    state_type=VimData,
    config=MyoComponent.cons(run_handler_for),
)

__all__ = ('vim',)



from amino import List, Maybe, Dat, ADT, Either, Map, Nil, Nothing, do, Do, Just

from myo.util import Ident
from myo.output.main import ParseConfig

from chiasma.util.id import IdentSpec, ensure_ident_or_generate, optional_ident, StrIdent


class Interpreter(ADT['Interpreter']):
    pass


class VimInterpreter(Interpreter):

    @staticmethod
    @do(Either[str, 'VimInterpreter'])
    def cons(silent: bool=False, target: IdentSpec=None) -> Do:
        ident = yield optional_ident(target)
        return VimInterpreter(silent, ident)

    @staticmethod
    def default() -> 'VimInterpreter':
        return VimInterpreter(False, Nothing)

    def __init__(self, silent: bool, target: Maybe[Ident]) -> None:
        self.silent = silent
        self.target = target


class SystemInterpreter(Interpreter):

    @staticmethod
    @do(Either[str, 'SystemInterpreter'])
    def cons(target: IdentSpec=None) -> Do:
        ident = yield optional_ident(target)
        return SystemInterpreter(ident)

    @staticmethod
    def str(target: str) -> 'SystemInterpreter':
        return SystemInterpreter(Just(StrIdent(target)))

    def __init__(self, target: Maybe[Ident]) -> None:
        self.target = target


class ShellInterpreter(Interpreter):

    @staticmethod
    @do(Either[str, 'ShellInterpreter'])
    def cons(target: IdentSpec) -> Do:
        ident = yield optional_ident(target)
        return ShellInterpreter(ident)

    @staticmethod
    def str(target: str) -> 'ShellInterpreter':
        return ShellInterpreter(StrIdent(target))

    def __init__(self, target: Ident) -> None:
        self.target = target


class CommandConfig(Dat['CommandConfig']):

    @staticmethod
    def cons(
            ident: IdentSpec,
            parsers: List[str]=None,
            output_filter: List[str]=None,
            output_first_error: str=None,
            output_path_formatter: str=None,
            output_reporter: str=None,
            output_syntax: str=None,
    ) -> 'CommandConfig':
        return CommandConfig(
            ensure_ident_or_generate(ident),
            Maybe.optional(parsers),
            Maybe.optional(output_filter),
            Maybe.optional(output_first_error),
            Maybe.optional(output_path_formatter),
            Maybe.optional(output_reporter),
            Maybe.optional(output_syntax),
        )

    def __init__(
            self,
            ident: Ident,
            parsers: Maybe[List[str]],
            output_filter: Maybe[List[str]],
            output_first_error: Maybe[str],
            output_path_formatter: Maybe[str],
            output_reporter: Maybe[str],
            output_syntax: Maybe[str],
    ) -> None:
        self.ident = ident
        self.parsers = parsers
        self.output_filter = output_filter
        self.output_first_error = output_first_error
        self.output_path_formatter = output_path_formatter
        self.output_reporter = output_reporter
        self.output_syntax = output_syntax

    @property
    def parse(self) -> ParseConfig:
        return ParseConfig(
            self.parsers.get_or_strict(Nil),
            self.output_filter.get_or_strict(Nil),
            self.output_first_error,
            self.output_path_formatter,
            self.output_reporter,
            self.output_syntax,
        )


class Command(Dat['Command']):

    @staticmethod
    def cons(
            ident: IdentSpec,
            interpreter: Interpreter=None,
            lines: List[str]=Nil,
            langs: List[str]=Nil,
            signals: List[str]=Nil,
            config: CommandConfig=None,
            history: bool=True,
    ) -> 'Command':
        return Command(
            ensure_ident_or_generate(ident),
            interpreter or SystemInterpreter(Nothing),
            lines,
            langs,
            signals,
            Maybe.optional(config),
            history
        )

    def __init__(
            self,
            ident: Ident,
            interpreter: Interpreter,
            lines: List[str],
            langs: List[str],
            signals: List[str],
            config: Maybe[CommandConfig],
            history: bool,
    ) -> None:
        self.ident = ident
        self.interpreter = interpreter
        self.lines = lines
        self.langs = langs
        self.signals = signals
        self.config = config
        self.history = history


class Execute(Dat['Execute']):

    def __init__(self, command: Ident) -> None:
        self.command = command


class TestLineParams(Dat['TestLineParams']):

    def __init__(
            self,
            line: str,
            shell: Either[str, str],
            target: Either[str, str],
            langs: List[str],
            options: Map[str, str],
    ) -> None:
        self.line = line
        self.shell = shell
        self.target = target
        self.langs = langs
        self.options = options


class TestCommand(Dat['TestCommand']):

    def __init__(self, command: Command, params: TestLineParams) -> None:
        self.command = command
        self.params = params


class HistoryEntry(Dat['HistoryEntry']):

    @staticmethod
    def cons(cmd: Command, target: Ident=None) -> 'HistoryEntry':
        return HistoryEntry(cmd, Maybe.optional(target))

    def __init__(self, cmd: Command, target: Maybe[Ident]) -> None:
        self.cmd = cmd
        self.target = target


class Pid(Dat['Pid']):

    def __init__(self, value: int) -> None:
        self.value = value


class RunningCommand(Dat['RunningCommand']):

    def __init__(self, ident: Ident, pid: Maybe[Pid], system: bool) -> None:
        self.ident = ident
        self.pid = pid
        self.system = system


__all__ = ('Command', 'Interpreter', 'VimInterpreter', 'SystemInterpreter', 'ShellInterpreter', 'Execute',
           'TestLineParams', 'TestCommand', 'HistoryEntry', 'RunningCommand', 'Pid', 'CommandConfig',)



from amino import do, Do, IO

from psutil import Process

from myo.data.command import Pid


def process(pid: Pid) -> IO[Process]:
    return IO.delay(Process, pid.value)


@do(IO[None])
def kill_process(pid: Pid, signal: int=9) -> Do:
    proc = yield process(pid)
    yield IO.delay(proc.kill)


__all__ = ('process', 'kill_process',)



from amino import do, Do
from amino.lenses.lens import lens

from chiasma.commands.pane import window_pane_open
from chiasma.window.main import window_by_ident
from chiasma.pane import pane_by_ident
from chiasma.io.state import TS
from chiasma.io.compute import TmuxIO

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo

from myo.ui.data.window import Window
from myo.ui.data.view import Pane
from myo.components.tmux.tpe import TmuxRibosome


@prog
@do(NS[TmuxRibosome, bool])
def tmux_window_pane_open(window: Window, pane: Pane) -> Do:
    native_window = yield Ribo.zoom_comp(window_by_ident(window.ident).nvim.zoom(lens.views))
    native_pane = yield Ribo.zoom_comp(pane_by_ident(pane.ident).nvim.zoom(lens.views))
    yield TS.lift((native_window.id & native_pane.id).map2(window_pane_open).get_or(TmuxIO.pure, False)).nvim


__all__ = ('tmux_window_pane_open',)



from ribosome.compute.api import prog
from ribosome.nvim.io.compute import NvimIO, NRParams
from ribosome.nvim.io.state import NS
from ribosome.nvim.api.command import nvim_command

from amino import do, Do, Boolean

from myo.env import Env
from myo.command.run_task import RunTask, VimTaskDetails


@prog.unit
@do(NS[Env, None])
def run(task: RunTask) -> Do:
    cmd = task.command
    params = NRParams.cons(verbose=not cmd.interpreter.silent, sync=False)
    yield NS.lift(cmd.lines.traverse(lambda a: nvim_command(a, params=params), NvimIO))
    yield NS.unit


def vim_can_run(task: RunTask) -> Boolean:
    return isinstance(task.details, VimTaskDetails)


__all__ = ('run', 'vim_can_run')



from typing import Callable, Any

from amino import Dat, Maybe, Nothing

from ribosome.config.component import Component, ComponentData
from ribosome.compute.program import Program

from myo.env import Env
from myo.ui.ui import Ui
from myo.command.run_task import RunTask
from myo.data.info import InfoWidget
from myo.data.command import Pid


def no_handler(*a: Any, **kw: Any) -> Maybe[Program]:
    return Nothing


class MyoComponent(Dat['MyoComponent']):

    @staticmethod
    def cons(
            run: Callable[[RunTask], Maybe[Program[Pid]]]=None,
            create_vim_pane: Callable[[], Maybe[Program]]=None,
            info: Program[InfoWidget]=None,
            ui: Ui=None,
            init: Program[None]=None,
            quit: Program[None]=None,
    ) -> 'MyoComponent':
        return MyoComponent(
            run or no_handler,
            create_vim_pane or no_handler,
            Maybe.optional(info),
            Maybe.optional(ui),
            Maybe.optional(init),
            Maybe.optional(quit),
        )

    def __init__(
            self,
            run: Callable[[RunTask], Maybe[Program[Pid]]],
            create_vim_pane: Callable[[RunTask], Maybe[Program]],
            info: Maybe[Program[InfoWidget]],
            ui: Maybe[Ui],
            init: Maybe[Program[None]],
            quit: Maybe[Program[None]],
    ) -> None:
        self.run = run
        self.create_vim_pane = create_vim_pane
        self.info = info
        self.ui = ui
        self.init = init
        self.quit = quit


Comp = Component[ComponentData, MyoComponent]

__all__ = ('MyoComponent',)



from myo.components.tmux.data import TmuxData
from chiasma.io.compute import TmuxIO
from chiasma.commands.pane import close_pane_id
from chiasma.io.state import TS

from amino import do, Do, _

from ribosome.nvim.io.state import NS
from ribosome.compute.api import prog
from ribosome.config.component import ComponentData

from myo.env import Env


@prog
@do(NS[ComponentData[Env, TmuxData], None])
def tmux_quit() -> Do:
    all_panes = yield NS.inspect(_.comp.views.panes)
    vim_pane_ident = yield NS.inspect(_.comp.vim_pane)
    close_panes = all_panes.filter(lambda a: not vim_pane_ident.contains(a.ident))
    yield TS.lift(close_panes.flat_map(_.id).traverse(close_pane_id, TmuxIO)).nvim


__all__ = ('tmux_quit',)



from amino import Dat


class CoreData(Dat['CoreData']):

    @staticmethod
    def cons() -> 'CoreData':
        return CoreData()

    def __init__(self) -> None:
        pass


__all__ = ('CoreData',)



from typing import Any, Tuple, Callable, Generic, TypeVar

from networkx import DiGraph

from amino import List, Maybe, curried, Just, _, Right, Either, Lists, Dat, Regex, Map, Nil, do, Do
from amino.regex import Match
from amino.func import tailrec
from amino.logging import module_log

from myo.output.data.output import OutputEvent, OutputLine

log = module_log()
A = TypeVar('A')
B = TypeVar('B')


class Parser(Generic[A, B], Dat['Parser[A, B]']):

    @staticmethod
    def cons(
            graph: DiGraph,
            cons_events: Callable[[List[OutputLine[A]]], List[OutputEvent[A, B]]],
    ) -> 'Parser':
        return Parser(
            graph,
            cons_events,
        )

    def __init__(self, graph: DiGraph, cons_events: Callable[[List[OutputLine[A]]], List[OutputEvent[A, B]]]) -> None:
        self.graph = graph
        self.cons_events = cons_events


class EdgeData(Generic[A], Dat['EdgeData[A]']):

    @staticmethod
    def strict(regex: Regex, cons_output_line: Callable[..., A]) -> 'EdgeData[A]':
        return EdgeData(regex, lambda *a, **kw: Right(cons_output_line(*a, **kw)))

    def __init__(self, regex: Regex, cons_output_line: Callable[..., Either[str, A]]) -> None:
        self.regex = regex
        self.cons_output_line = cons_output_line


class Step(Dat['Step']):

    def __init__(self, node: str, data: Map, weight: int) -> None:
        self.node = node
        self.data = data
        self.weight = weight


@curried
def cons_output_line(cons: Callable[..., Either[str, A]]) -> Callable[[Match], Either[str, OutputLine[A]]]:
    @do(Either[str, OutputLine[A]])
    def cons_output_line(match: Match) -> Do:
        data = yield cons(**match.group_map)
        return OutputLine(match.string, data)
    return cons_output_line


def simple_event(lines: List[OutputLine[A]]) -> Maybe[OutputEvent[A, B]]:
    return Just(OutputEvent(lines=lines))


def edges(graph: DiGraph, node: str) -> List[Step]:
    e = (Step(t, d['data'], d.get('weight', 0)) for f, t, d in graph.edges(node, data=True))
    return Lists.wrap(e)


def match_edge(
        graph: DiGraph,
        node: str,
        match: Callable[[Any], Either[str, OutputLine[A]]],
) -> Either[str, OutputLine[A]]:
    return (
        edges(graph, node)
        .sort_by(_.weight, reverse=True)
        .find_map_e(match)
    )


@tailrec
def simple_parse_process(
        parser: Parser[A, B],
        node: str,
        output: List[str],
        result: List[OutputEvent[A, B]],
        current: List[OutputLine[A]],
) -> List[OutputEvent[A, B]]:
    '''
    Parse a list of output lines.
    The algorithm starts at the graph node 'start'
    1. detach the first output line into *line* and *rest* and call
        *parse_line*
    2. find an edge that matches to current line
    a) if a match was found
        3. construct an *OutputLine*
        4. set the current node to the target node of the edge
        5. add the entry to the list *current*
        6. recurse with *rest* as new *output*
    b) if no match was found
        7. construct an *OutputEvent* from *current*
        8. if the current node is 'start', set *output* to *rest*
            else, keep *output* to try again with 'start'
        9. recurse with 'start'
    10. add the last event and exit the recursion
    '''
    def add_event() -> List[OutputEvent[A, B]]:
        new = Nil if current.empty else parser.cons_events(current)
        return result + new
    def parse_line(line: str, rest: List[str]) -> Tuple[bool, tuple]:
        log.debug2(lambda: f'parsing line: {line}')
        @do(Either[str, Tuple[OutputLine[A], str]])
        def match(step: Step) -> Do:
            match = yield step.data.regex.match(line)
            log.debug2(lambda: f'matched edge to {step.node}')
            output_line = yield cons_output_line(step.data.cons_output_line)(match)
            return output_line, step.node
        def cont(entry: OutputLine[A], next_node: str) -> Tuple[bool, tuple]:
            return True, (parser, next_node, rest, result, current.cat(entry))
        def next_event() -> Tuple[bool, tuple]:
            new_output = rest if node == 'start' and current.empty else output
            return True, (parser, 'start', new_output, add_event(), List())
        return match_edge(parser.graph, node, match).map2(cont) | next_event
    quit = lambda: (False, add_event())
    return output.detach_head.map2(parse_line) | quit


def parse_events(parser: Parser[A, B], output: List[str]) -> Either[str, List[OutputEvent[A, B]]]:
    log.debug(f'parsing with {parser}')
    return Right(simple_parse_process(parser, 'start', output, Nil, Nil))


__all__ = ('parse_events', 'Parser', 'EdgeData', 'Step', 'cons_output_line',)



from typing import Union

from networkx import DiGraph

from amino import List, Just, Maybe, Path, Regex, Nothing, do, Either, Do, Try, ADT, Nil, Dat
from amino.util.numeric import parse_int
from amino.case import Case
from amino.logging import module_log

from myo.output.data.output import OutputLine, OutputEvent, Location
from myo.output.parser.base import EdgeData, Parser

log = module_log()


class PythonLine(ADT['PythonLine']):
    pass


class CodeLine(PythonLine):

    def __init__(self, code: str) -> None:
        self.code = code


class FileLine(PythonLine):

    @staticmethod
    @do(Either[str, 'FileLine'])
    def cons(
            path: Union[str, Path],
            line: Union[str, int],
            fun: str=None,
    ) -> Do:
        path_p = yield Try(Path, path)
        line_i = yield parse_int(line)
        return FileLine(path_p, line_i - 1, Maybe.optional(fun))

    def __init__(self, path: Path, line: int, fun: Maybe[str]) -> None:
        self.path = path
        self.line = line
        self.fun = fun


class ErrorLine(PythonLine):

    def __init__(self, error: str, exc: str) -> None:
        self.error = error
        self.exc = exc


class ColLine(PythonLine):

    @staticmethod
    def whitespace(ws: str) -> 'ColLine':
        return ColLine(len(ws))

    def __init__(self, col: int) -> None:
        self.col = col


file_edge = EdgeData(
    regex=Regex('\s*File "(?P<path>.+)", line (?P<line>\d+)(?:, in (?P<fun>\S+))?'),
    cons_output_line=FileLine.cons
)
code_edge = EdgeData.strict(regex=Regex('^\s*(?P<code>.+)'), cons_output_line=CodeLine)
error_edge = EdgeData.strict(regex=Regex('^\s*(?P<exc>\S+): (?P<error>.+)'), cons_output_line=ErrorLine)
col_edge = EdgeData.strict(regex=Regex('^\s*(?P<ws>\s+)\^'), cons_output_line=ColLine.whitespace)


def python_graph() -> DiGraph:
    g = DiGraph()
    g.add_edge('start', 'file', data=file_edge)
    g.add_edge('file', 'code', data=code_edge)
    g.add_edge('code', 'file', data=file_edge)
    g.add_edge('code', 'error', data=error_edge)
    g.add_edge('code', 'col', data=col_edge)
    g.add_edge('col', 'error', data=error_edge)
    return g


class EventData(Dat['EventData']):

    @staticmethod
    def cons(
            output: List[OutputEvent]=Nil,
            file: Maybe[FileLine]=Nothing,
            code: Maybe[CodeLine]=Nothing,
            col: Maybe[ColLine]=Nothing,
    ) -> 'EventData':
        return EventData(
            output,
            file,
            code,
            col,
        )

    def __init__(
            self,
            output: List[OutputEvent],
            file: Maybe[FileLine],
            code: Maybe[CodeLine],
            col: Maybe[ColLine],
    ) -> None:
        self.output = output
        self.file = file
        self.code = code
        self.col = col


class PythonEvent(ADT['PythonEvent']):
    pass


class StackFrameEvent(PythonEvent):

    def __init__(self, file: OutputLine[FileLine], code: Maybe[OutputLine[CodeLine]]) -> None:
        self.file = file
        self.code = code


class ErrorEvent(PythonEvent):

    def __init__(self, line: OutputLine[PythonLine]) -> None:
        self.line = line


def stack_frame_event(
        file: OutputLine[FileLine],
        code: Maybe[OutputLine[CodeLine]],
        col: Maybe[OutputLine[ColLine]],
) -> OutputEvent:
    lines = List(file).cat_m(code)
    location = Location(file.meta.path, file.meta.line, col.map(lambda a: a.meta.col).get_or_strict(0))
    return OutputEvent.cons(StackFrameEvent(file, code), lines, Just(location))


def push_event(data: EventData) -> EventData:
    event = data.file.map(lambda a: stack_frame_event(a, data.code, data.col))
    new = event.cata(data.append1.output, data)
    return EventData.cons(output=new.output)


class update_event_data(Case[PythonLine, EventData], alg=PythonLine):

    def __init__(self, current: EventData, line: OutputLine[PythonLine]) -> None:
        self.current = current
        self.line = line

    def error(self, meta: ErrorLine) -> EventData:
        return push_event(self.current).append1.output(OutputEvent.cons(ErrorEvent(self.line), List(self.line)))

    def code(self, meta: CodeLine) -> EventData:
        return self.current.set.code(Just(self.line))

    def col(self, meta: ColLine) -> EventData:
        return self.current.set.col(Just(self.line))

    def file(self, meta: FileLine) -> EventData:
        return push_event(self.current).set.file(Just(self.line))


def python_events(lines: List[OutputLine[PythonLine]]) -> List[OutputEvent[PythonLine, PythonEvent]]:
    data = lines.fold_left(EventData.cons())(lambda z, a: update_event_data(z, a)(a.meta))
    return push_event(data).output


python_parser = Parser(python_graph(), python_events)

__all__ = ('python_parser',)



from typing import TypeVar, Callable, Tuple

from amino import do, Do, List, Just, Boolean, IO, Either, Left, Lists, Nothing
from amino.boolean import true, false
from amino.lenses.lens import lens
from amino.logging import module_log
from amino.case import Case
from amino.state import State

from ribosome.nvim.io.state import NS
from ribosome.nvim.scratch import show_in_scratch_buffer_default, ScratchBuffer
from ribosome.compute.api import prog
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.ui import (window_line, focus_window, edit_file, set_local_cursor, set_line, window_buffer_name,
                                  close_buffer, window_number, close_window)
from ribosome.nvim.io.api import N
from ribosome.components.internal.mapping import activate_mapping
from ribosome.data.mapping import Mapping
from ribosome.compute.ribosome_api import Ribo
from ribosome.nvim.syntax.cmd import syntax_item_cmd, highlight_cmd, hi_link_cmd
from ribosome.nvim.api.command import nvim_atomic_commands, nvim_command
from ribosome.nvim.api.util import format_windo
from ribosome import ribo_log

from myo.components.command.data import CommandData, OutputData
from myo.output.data.output import Location, OutputEvent
from myo.components.command.compute.tpe import CommandRibosome
from myo.config.settings import auto_jump
from myo.output.data.report import (ReportLine, format_report, ParseReport, PlainReportLine, EventReportLine,
                                    event_index, DisplayLine)
from myo.components.command.compute.parsed_output import ParsedOutput
from myo.components.command.compute.parse_handlers import Reporter, SyntaxCons

log = module_log()
A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')
jump_mapping = Mapping.cons('output_jump', '<cr>', true)
quit_mapping = Mapping.cons('output_quit', 'q', true)
prev_mapping = Mapping.cons('output_prev', '<m-->', false)
next_mapping = Mapping.cons('output_next', '<m-=>', false)


def output_data() -> NS[CommandData, OutputData]:
    return NS.inspect(lambda a: a.output)


def scratch_buffer() -> NS[CommandData, ScratchBuffer]:
    return NS.inspect_either(lambda a: a.output.scratch.to_either('no scratch buffer set'))


def report_line_number(event: int) -> Callable[[ParseReport], Either[str, int]]:
    err = lambda: f'no matching line for event {event}'
    def report_line_number(report: ParseReport) -> Either[str, int]:
        return report.event_indexes.lift(event).to_either_f(err)
    return report_line_number


@do(NS[CommandData, int])
def inspect_report_line_number_by_event_index(index: int) -> Do:
    yield (
        NS.inspect_either(report_line_number(index))
        .zoom(lens.output.report)
    )


@do(NS[CommandData, int])
def line_event_index(line: int) -> Do:
    indexes = yield NS.inspect(lambda a: a.output.report.event_indexes)
    return indexes.index_where(lambda a: a > line + 1).get_or_strict(len(indexes)) - 1


no_associated_location = 'no location associated with this event'


class report_line_location(Case[ReportLine, Either[str, Location]], alg=ReportLine):

    def plain(self, line: PlainReportLine) -> Either[str, Location]:
        return Left(no_associated_location)

    def event(self, line: EventReportLine) -> Either[str, Location]:
        return line.location.to_either(no_associated_location)


def line_location(line: int) -> Callable[[ParseReport], Either[str, Location]]:
    @do(Either[str, Location])
    def report_line_number(report: ParseReport) -> Do:
        report_line = yield report.lines.lift(line).to_either(lambda: f'invalid report line index {line}')
        yield report_line_location.match(report_line)
    return report_line_number


@do(NS[CommandData, Location])
def inspect_line_location(line: int) -> Do:
    yield NS.inspect_either(line_location(line)).zoom(lens.output.report)


@do(NvimIO[None])
def jump_to_location(scratch: ScratchBuffer, location: Location) -> Do:
    window = scratch.ui.previous
    yield focus_window(window)
    current_file = yield window_buffer_name(window)
    location_exists = yield N.from_io(IO.delay(location.path.exists))
    if location_exists:
        yield edit_file(location.path) if current_file != str(location.path) else N.unit
        yield set_local_cursor(location.coords)
        yield nvim_command('normal! zvzz')
    yield N.unit


def store_event_index(index: int) -> NS[CommandData, None]:
    return NS.s(CommandData).modify(lambda a: a.set.current(index)).zoom(lens.output)


@do(NS[CommandData, None])
def select_event(index: int, jump: Boolean) -> Do:
    yield store_event_index(index)
    line = yield inspect_report_line_number_by_event_index(index)
    scratch = yield scratch_buffer()
    yield NS.lift(set_line(scratch.ui.window, line))
    if jump:
        location = yield inspect_line_location(line)
        yield NS.lift(jump_to_location(scratch, location))


@do(NS[CommandData, None])
def select_line_event(line: int) -> Do:
    index = yield line_event_index(line)
    yield store_event_index(index)


def event_report(index: int, event: OutputEvent[A, B], lines: List[DisplayLine]) -> List[ReportLine]:
    return lines.map(lambda a: EventReportLine(a, index, event.location))


@do(NS[CommandRibosome, List[List[ReportLine]]])
def events_report(output: ParsedOutput[A, B], reporter: Reporter) -> Do:
    reports = yield output.filtered.traverse(lambda a: reporter(output.handlers.path_formatter, a), NS)
    return Lists.range(len(reports)).zip(output.filtered, reports).map3(event_report)


@do(NS[CommandRibosome, ParseReport])
def parse_report(output: ParsedOutput[A, B]) -> Do:
    events_lines = yield events_report(output, output.handlers.reporter)
    lines = events_lines.join
    event_indexes = events_lines.traverse(event_index, State).run_a(0).value
    return ParseReport(lines, event_indexes)


@do(NS[CommandRibosome, None])
def setup_syntax(cons: SyntaxCons, window: int) -> Do:
    syntax = yield cons()
    cmds = (
        syntax.syntax.map(syntax_item_cmd.match) +
        syntax.highlight.map(highlight_cmd) +
        syntax.links.map(hi_link_cmd) +
        List('setlocal conceallevel=2')
    )
    win_cmds = cmds.map(lambda a: format_windo(a, window, Nothing))
    yield NS.lift(nvim_atomic_commands(win_cmds))


@do(NvimIO[None])
def close_scratch_buffer(scratch: ScratchBuffer) -> Do:
    yield N.ignore_failure(close_window(scratch.ui.window))
    yield N.ignore_failure(close_buffer(scratch.buffer))


@do(NS[CommandRibosome, None])
def terminate_scratch() -> Do:
    scratch = yield Ribo.inspect_comp(lambda a: a.output.scratch)
    yield NS.lift(scratch.cata(close_scratch_buffer, N.unit))


@do(NS[CommandRibosome, None])
def empty_report() -> Do:
    ribo_log.error('no events in command output')
    yield NS.unit


@do(NS[CommandRibosome, None])
def render_report(output: ParsedOutput[A, B], report: ParseReport) -> Do:
    yield terminate_scratch()
    scratch = yield NS.lift(show_in_scratch_buffer_default(format_report(report), max_height=Just(.3)))
    scratch_number = yield NS.lift(window_number(scratch.ui.window))
    yield Ribo.modify_comp(lens.output.modify(lambda a: a.copy(report=report, scratch=Just(scratch))))
    yield List(jump_mapping, quit_mapping, prev_mapping, next_mapping).traverse(activate_mapping, NS).zoom(lens.state)
    yield setup_syntax(output.handlers.syntax, scratch_number)
    jump = yield Ribo.setting(auto_jump)
    first_error = yield output.handlers.first_error(output.filtered)
    yield Ribo.zoom_comp(select_event(first_error, jump))


@do(NS[CommandRibosome, None])
def render_parse_result(output: ParsedOutput[A, B]) -> Do:
    log.debug(f'rendering parse result')
    report = yield parse_report(output)
    yield empty_report() if report.event_indexes.empty else render_report(output, report)


@do(NS[CommandRibosome, Tuple[ScratchBuffer, int]])
def select_cursor_event() -> Do:
    scratch = yield Ribo.zoom_comp(scratch_buffer())
    line = yield NS.lift(window_line(scratch.ui.window))
    yield Ribo.zoom_comp(select_line_event(line))
    return scratch, line


@prog
@do(NS[CommandRibosome, None])
def current_event_jump() -> Do:
    scratch, line = yield select_cursor_event()
    location = yield Ribo.zoom_comp(inspect_line_location(line))
    yield NS.lift(jump_to_location(scratch, location))


@prog.comp
@do(NS[CommandData, None])
def quit_output() -> Do:
    scratch = yield scratch_buffer()
    yield NS.lift(close_buffer(scratch.buffer))


@prog
@do(NS[CommandRibosome, None])
def prev_event() -> Do:
    current = yield Ribo.inspect_comp(lambda a: a.output.current)
    if current > 0:
        jump = yield Ribo.setting(auto_jump)
        yield Ribo.zoom_comp(select_event(current - 1, jump))


@prog
@do(NS[CommandRibosome, None])
def next_event() -> Do:
    current = yield Ribo.inspect_comp(lambda a: a.output.current)
    count = yield Ribo.inspect_comp(lambda a: a.output.report.event_indexes.length)
    if current < count - 1:
        jump = yield Ribo.setting(auto_jump)
        yield Ribo.zoom_comp(select_event(current + 1, jump))


__all__ = ('render_parse_result', 'current_event_jump', 'quit_output', 'prev_event', 'next_event')



from amino import List
from amino.logging import module_log

from myo.output.data.output import OutputEvent

from ribosome.nvim.io.state import NS
from myo.output.parser.python import PythonLine, PythonEvent, ErrorEvent
from myo.components.command.compute.tpe import CommandRibosome

log = module_log()


def python_first_error(output: List[OutputEvent[PythonLine, PythonEvent]]) -> NS[CommandRibosome, int]:
    index = output.index_where(lambda a: isinstance(a.meta, ErrorEvent)).map(lambda a: abs(a - 1))
    return NS.pure(index.get_or_strict(0))


__all__ = ('python_first_error',)



from typing import Tuple

from amino import List, do, Do, Nil, Maybe, Lists, Dat, Eval, Just
from amino.util.string import indent
from amino.logging import module_log
from amino.util.numeric import add, sub

from myo.components.command.compute.tpe import CommandRibosome
from myo.output.data.output import OutputEvent, Location, OutputLine
from myo.components.command.compute.parse_handlers import PathFormatter
from myo.output.data.report import DisplayLine, PlainDisplayLine
from myo.output.parser.scala import ScalaEvent, ScalaLine, ColLine, CodeLine

from ribosome.nvim.io.state import NS

log = module_log()
col_marker = '†'
foundreq_separator = '|'
found_marker = '+'
separator_marker = '<'
req_marker = '-'


@do(NS[CommandRibosome, List[DisplayLine]])
def format_location(location: Location, path_formatter: PathFormatter) -> Do:
    path = yield path_formatter(location.path)
    return List(PlainDisplayLine(f'{path}  {location.line + 1}'))


def inject_col_marker(code_line: CodeLine, col_line: Maybe[OutputLine[ColLine]]) -> str:
    col = col_line.map(lambda a: a.meta.col).get_or_strict(0) - code_line.indent
    return f'{code_line.code[:col]}{col_marker}{code_line.code[col:]}'


class BracketState(Dat['BracketState']):

    @staticmethod
    def cons(
            match: int=None,
            index: int=0,
            parens: int=0,
            square: int=0,
    ) -> 'BracketState':
        return BracketState(
            Maybe.optional(match),
            index,
            parens,
            square,
        )

    def __init__(self, match: Maybe[int], index: int, parens: int, square: int) -> None:
        self.match = match
        self.index = index
        self.parens = parens
        self.square = square

    def push_square(self) -> 'BracketState':
        return self.mod.square(add(1))

    def push_parens(self) -> 'BracketState':
        return self.mod.parens(add(1))

    def pop_square(self) -> 'BracketState':
        return self.set.match(Just(self.index)) if self.square == 0 else self.mod.square(sub(1))

    def pop_parens(self) -> 'BracketState':
        return self.set.match(Just(self.index)) if self.parens == 0 else self.mod.parens(sub(1))


def limiting_bracket(chars: List[str], brackets: Tuple[str, str, str, str]) -> Maybe[int]:
    o_square, c_square, o_parens, c_parens = brackets
    def folder(z: BracketState, a: str) -> Eval[BracketState]:
        def step() -> BracketState:
            next = (
                z.push_square()
                if a == c_square else
                z.push_parens()
                if a == c_parens else
                z.pop_square()
                if a == o_square else
                z.pop_parens()
                if a == c_parens else
                z
            )
            return next.mod.index(add(1))
        return Eval.now(z.match.replace(z).get_or(step))
    result = chars.fold_m(Eval.now(BracketState.cons()))(folder)
    return result.evaluate().match


def opening_bracket(line: str, index: int) -> int:
    chars = Lists.wrap(line[:index]).reversed
    result = limiting_bracket(chars, ('[', ']', '(', ')'))
    return result.map(lambda a: index - a).get_or_strict(0)


def closing_bracket(line: str, index: int) -> int:
    chars = Lists.wrap(line[index:])
    result = limiting_bracket(chars, (']', '[', ')', '('))
    return result.map(lambda a: index + a).get_or_strict(len(line))


def foundreq_boundaries(line: str, index: int) -> Tuple[int, int, int]:
    opening = opening_bracket(line, index)
    closing = closing_bracket(line, index)
    return opening, index, closing


def insert_bound_markers(line: str, bounds: Tuple[int, int, int]) -> str:
    start, separator, end = bounds
    left = f'{line[:start]}{found_marker}{line[start:separator]}{separator_marker}'
    right = f'{line[separator:end]}{req_marker}{line[end:]}'
    return left + right


def handle_foundreq(line: str) -> str:
    indexes = Lists.find_str_matches(line, foundreq_separator).reversed
    bounds = indexes.map(lambda a: foundreq_boundaries(line, a))
    return bounds.fold_left(line)(insert_bound_markers)


@do(NS[CommandRibosome, List[DisplayLine]])
def scala_report(path_formatter: PathFormatter, output: OutputEvent[ScalaLine, ScalaEvent]) -> Do:
    code = (
        output.meta.code
        .map(lambda a: indent(inject_col_marker(a.meta, output.meta.col)))
        .map(PlainDisplayLine)
    )
    info = output.meta.info.map(lambda a: handle_foundreq(a.meta.message)).indent(2).map(PlainDisplayLine)
    location = yield output.location.map(lambda a: format_location(a, path_formatter)).get_or(NS.pure, Nil)
    error = PlainDisplayLine(output.meta.file.meta.error)
    return location.cat(error) + info + code.to_list


__all__ = ('scala_report',)



from amino import Do, Maybe, Nothing
from amino.case import Case

from ribosome.compute.api import prog

from chiasma.util.id import IdentSpec, Ident
from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo

from myo.util.id import ensure_ident_ns
from myo.components.command.data import CommandData
from myo.components.command.compute.run import command_by_ident, run_command
from myo.data.command import Command, Interpreter, SystemInterpreter
from myo.components.ui.compute.close_pane import close_pane


class target_pane(Case[Interpreter, Maybe[Ident]], alg=Interpreter):

    def system_interpreter(self, interpreter: SystemInterpreter) -> Maybe[Ident]:
        return interpreter.target

    def case_default(self, interpreter: Interpreter) -> Maybe[Ident]:
        return Nothing


@prog.do(None)
def reboot_command_pane(cmd: Command) -> Do:
    pane = target_pane.match(cmd.interpreter)
    yield pane.cata(close_pane, NS.unit)
    yield run_command(cmd)


@prog.do(None)
def reboot(ident_spec: IdentSpec) -> Do:
    ident = yield Ribo.lift_comp(ensure_ident_ns(ident_spec), CommandData)
    cmd = yield command_by_ident(ident)
    yield reboot_command_pane(cmd)


__all__ = ('reboot',)



from amino import Dat, Maybe, List, Nil
from amino.tc.monoid import Monoid


class ParseConfig(Dat['ParseConfig']):

    @staticmethod
    def cons(
            langs: List[str]=Nil,
            filter: List[str]=Nil,
            first_error: str=None,
            path_formatter: str=None,
            reporter: str=None,
            syntax: str=None,
    ) -> 'ParseConfig':
        return ParseConfig(
            langs,
            filter,
            Maybe.optional(first_error),
            Maybe.optional(path_formatter),
            Maybe.optional(reporter),
            Maybe.optional(syntax),
        )

    def __init__(
            self,
            langs: List[str],
            filter: List[str],
            first_error: Maybe[str],
            path_formatter: Maybe[str],
            reporter: Maybe[str],
            syntax: Maybe[str],
    ) -> None:
        self.langs = langs
        self.filter = filter
        self.first_error = first_error
        self.path_formatter = path_formatter
        self.reporter = reporter
        self.syntax = syntax


class Monoid_ParseConfig(Monoid, tpe=ParseConfig):

    @property
    def empty(self) -> ParseConfig:
        return ParseConfig.cons()

    def combine(self, a: ParseConfig, b: ParseConfig) -> ParseConfig:
        return ParseConfig(
            a.langs + b.langs,
            a.filter + b.filter,
            a.first_error.o(b.first_error),
            a.path_formatter.o(b.path_formatter),
            a.reporter.o(b.reporter),
            a.syntax.o(b.syntax),
        )


class LangConfig(Dat['LangConfig']):

    @staticmethod
    def cons(
            name: str,
            output_filter: List[str]=None,
            output_first_error: str=None,
            output_path_formatter: str=None,
            output_reporter: str=None,
            output_syntax: str=None,
    ) -> 'LangConfig':
        return LangConfig(
            name,
            Maybe.optional(output_filter),
            Maybe.optional(output_first_error),
            Maybe.optional(output_path_formatter),
            Maybe.optional(output_reporter),
            Maybe.optional(output_syntax),
        )

    def __init__(
            self,
            name: str,
            output_filter: Maybe[List[str]],
            output_first_error: Maybe[str],
            output_path_formatter: Maybe[str],
            output_reporter: Maybe[str],
            output_syntax: Maybe[str],
    ) -> None:
        self.name = name
        self.output_filter = output_filter
        self.output_first_error = output_first_error
        self.output_path_formatter = output_path_formatter
        self.output_reporter = output_reporter
        self.output_syntax = output_syntax

    @property
    def parse(self) -> ParseConfig:
        return ParseConfig(
            Nil,
            self.output_filter.get_or_strict(Nil),
            self.output_first_error,
            self.output_path_formatter,
            self.output_reporter,
            self.output_syntax,
        )


__all__ = ('ParseConfig', 'LangConfig',)



from ribosome.compute.ribosome import Ribosome

from myo.env import Env
from myo.config.component import MyoComponent
from myo.components.command.data import CommandData

CommandRibosome = Ribosome[Env, MyoComponent, CommandData]


__all__ = ('CommandRibosome',)



from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo

from chiasma.util.id import IdentSpec, Ident

from amino import do, Do, Just, Dat, Maybe, List
from amino.logging import module_log

from myo.components.command.compute.output import render_parse_result
from myo.components.command.compute.tpe import CommandRibosome
from myo.config.settings import display_parse_result
from myo.components.command.parse import parse_most_recent

log = module_log()


class ParseOptions(Dat['ParseOptions']):

    @staticmethod
    def cons(pane: IdentSpec=None, command: IdentSpec=None, langs: List[str]=None) -> 'ParseOptions':
        return ParseOptions(Maybe.optional(pane), Maybe.optional(command), Maybe.optional(langs))

    def __init__(self, pane: Maybe[Ident], command: Maybe[Ident], langs: Maybe[List[str]]) -> None:
        self.pane = pane
        self.command = command
        self.langs = langs


@prog.unit
@do(NS[CommandRibosome, None])
def parse(options: ParseOptions) -> Do:
    parse_result = yield parse_most_recent()
    yield Ribo.modify_comp(lambda a: a.set.parse_result(Just(parse_result)))
    display_result = yield Ribo.setting(display_parse_result)
    yield render_parse_result(parse_result) if display_result else NS.unit


__all__ = ('parse',)



from typing import TypeVar, Generic

from amino import Dat, List, do, Do, Maybe, Boolean, L, _, Nil, Nothing
from amino.case import Case
from amino.state import EitherState

from chiasma.data.view_tree import ViewTree, LayoutNode, PaneNode, SubUiNode

from myo.ui.data.view import Layout
from myo.ui.data.space import Space
from myo.ui.data.window import Window
from myo.util import Ident
from myo.ui.data.ui_data import UiData

A = TypeVar('A')


class ViewPath(Generic[A], Dat['ViewPath']):

    @staticmethod
    def cons(
        view: A,
        layouts: List[Layout],
        window: Window,
        space: Space,
    ) -> 'ViewPath':
        return ViewPath(
            view,
            layouts,
            window,
            space,
        )

    def __init__(self, view: A, layouts: List[Layout], window: Window, space: Space) -> None:
        self.view = view
        self.layouts = layouts
        self.window = window
        self.space = space

    @property
    def root(self) -> ViewTree:
        return self.window.layout


class pane_path_view(Case, alg=ViewTree):

    def __init__(self, ident: Ident, stack: List[Layout], window: Window, space: Space) -> None:
        self.ident = ident
        self.stack = stack
        self.window = window
        self.space = space

    def recurse(self, layout: Layout, node: ViewTree) -> Maybe[ViewPath]:
        return pane_path_view(self.ident, self.stack.cons(layout), self.window, self.space)(node)

    def layout_node(self, node: LayoutNode) -> Maybe[ViewPath]:
        return node.sub.find_map(L(self.recurse)(node.data, _))

    def pane_node(self, node: PaneNode) -> Maybe[ViewPath]:
        pane = node.data
        return Boolean(pane.ident == self.ident).m(lambda: ViewPath(pane, self.stack, self.window, self.space))

    def sub_ui_node(self, node: SubUiNode) -> Maybe[ViewPath]:
        return Nothing


def pane_path_space(ident: Ident, space: Space) -> Maybe[ViewPath]:
    return space.windows.find_map(lambda a: pane_path_view(ident, Nil, a, space)(a.layout))


@do(EitherState[str, UiData, ViewPath])
def pane_path(ident: Ident) -> Do:
    spaces = yield EitherState.inspect(_.spaces)
    path_m = spaces.find_map(L(pane_path_space)(ident, _))
    yield EitherState.lift(path_m.to_either(f'pane not found: {ident}'))


__all__ = ('ViewPath', 'pane_path')



from typing import Union

from amino import do, Do

from chiasma.util.id import ensure_ident_or_generate
from amino.boolean import true
from amino.state import EitherState
from amino.lenses.lens import lens

from ribosome.compute.api import prog
from ribosome.config.component import ComponentData
from ribosome.compute.prog import Prog
from ribosome.nvim.io.state import NS

from myo.util import Ident
from myo.ui.data.ui_data import UiData
from myo.ui.data.window import Window
from myo.env import Env
from myo.ui.pane import ui_modify_pane
from myo.components.ui.compute.pane import render_view
from myo.util.id import ensure_ident_prog


def ui_minimize_pane(ident: Ident) -> EitherState[str, UiData, Window]:
    return ui_modify_pane(ident, lens.state.minimized.set(true))


@prog
@do(NS[ComponentData[Env, UiData], Window])
def ui_minimize_pane_prog(ident: Ident) -> Do:
    yield ui_minimize_pane(ident).nvim.zoom(lens.comp)


@prog.do(None)
def minimize_pane(ident_spec: Union[str, Ident]) -> Do:
    ident = yield ensure_ident_prog(ident_spec)
    yield ui_minimize_pane_prog(ident)
    yield render_view(ident)


__all__ = ('minimize_pane',)



from amino import do, Do, Dat, Maybe, Nothing, _

from chiasma.util.id import Ident
from chiasma.ui.view_geometry import ViewGeometry
from chiasma.ui.state import ViewState

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo

from myo.ui.data.ui_data import UiData
from myo.ui.data.view import Pane
from myo.ui.pane import insert_pane_into_ui
from myo.components.ui.compute.tpe import UiRibosome


class CreatePaneOptions(Dat['CreatePaneOptions']):

    @staticmethod
    def cons(
            layout: Ident,
            ident: Maybe[Ident]=Nothing,
            min_size: Maybe[int]=Nothing,
            max_size: Maybe[int]=Nothing,
            fixed_size: Maybe[int]=Nothing,
            minimized_size: Maybe[int]=Nothing,
            weight: Maybe[float]=Nothing,
            position: Maybe[int]=Nothing,
            minimized: Maybe[bool]=False,
    ) -> 'CreatePaneOptions':
        return CreatePaneOptions(
            layout,
            ident,
            min_size,
            max_size,
            fixed_size,
            minimized_size,
            weight,
            position,
            minimized,
        )

    def __init__(
            self,
            layout: Ident,
            ident: Maybe[Ident],
            min_size: Maybe[int],
            max_size: Maybe[int],
            fixed_size: Maybe[int],
            minimized_size: Maybe[int],
            weight: Maybe[float],
            position: Maybe[int],
            minimized: Maybe[bool],
    ) -> None:
        self.layout = layout
        self.ident = ident
        self.min_size = min_size
        self.max_size = max_size
        self.fixed_size = fixed_size
        self.minimized_size = minimized_size
        self.weight = weight
        self.position = position
        self.minimized = minimized


def pane_from_options(options: CreatePaneOptions) -> Pane:
    return Pane.cons(
        ident=options.ident | Ident.generate,
        state=ViewState.cons(options.minimized | False),
        geometry=ViewGeometry(
            options.min_size,
            options.max_size,
            options.fixed_size,
            options.minimized_size,
            options.weight,
            options.position,
        )
    )


@do(NS[UiData, None])
def insert_pane(pane: Pane, layout: Ident) -> Do:
    updated = yield NS.inspect_either(insert_pane_into_ui(pane, layout))
    yield NS.set(updated)


@prog.unit
@do(NS[UiRibosome, None])
def create_pane(options: CreatePaneOptions) -> Do:
    pane = pane_from_options(options)
    yield Ribo.zoom_comp(insert_pane(pane, options.layout))


__all__ = ('create_pane',)



from amino import Do

from chiasma.util.id import IdentSpec, ensure_ident_or_generate

from ribosome.compute.api import prog
from ribosome.compute.prog import Prog

from myo.components.ui.compute.pane import view_owners
from myo.util.id import ensure_ident_prog


@prog.do(None)
def kill_pane(ident_spec: IdentSpec) -> Do:
    ident = yield ensure_ident_prog(ident_spec)
    space, window, owner = yield view_owners(ident)
    kill_handler = yield Prog.m(owner.kill_pane, lambda: f'ui `{type(owner)}` has no handler for `kill_pane`')
    yield kill_handler(ident)


__all__ = ('kill_pane',)



from amino import do, Do, Maybe, IO, Lists

from chiasma.io.compute import TmuxIO

from psutil import Process
from chiasma.commands.pane import pane_pid

from myo.data.command import Pid


@do(TmuxIO[Maybe[Pid]])
def process_pid(pane_id: int) -> Do:
    shell_pid = yield pane_pid(pane_id)
    proc = yield TmuxIO.from_io(IO.delay(Process, shell_pid))
    children = yield TmuxIO.from_io(IO.delay(proc.children))
    return Lists.wrap(children).head.map(lambda a: a.pid).map(Pid)


__all__ = ('process_pid',)



from amino import Map, Maybe, __
from amino.util.numeric import parse_int

from chiasma.util.id import Ident


def parse_id(value, rex, desc):
    return (
        Maybe(rex.match(str(value)))
        .map(__.group(1))
        .map(int)
        .to_either("could not match {} id {}".format(desc, value))
        .or_else(lambda: parse_int(value)))


def amend_options(opt: Map, key: str, value: Maybe):
    return value / (lambda a: opt + (key, a)) | opt


__all__ = ('parse_int', 'parse_id', 'Ident')



from amino import do, Do, __, List
from amino.logging import module_log

from chiasma.util.id import StrIdent, Ident
from chiasma.data.view_tree import ViewTree

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.function import nvim_call_tpe
from ribosome.compute.ribosome_api import Ribo
from ribosome.compute.prog import Prog
from ribosome.compute.program import Program
from ribosome import ribo_log

from myo.ui.data.ui_data import UiData
from myo.ui.data.space import Space
from myo.config.handler import find_handler_e
from myo.ui.data.view import Pane, Layout
from myo.ui.data.window import Window
from myo.config.settings import init_default_ui, vim_pane_geometry

log = module_log()


def vim_pid() -> NvimIO[int]:
    return nvim_call_tpe(int, 'getpid')


@do(NS[UiData, None])
def insert_default_ui(ident: Ident, layout_ident: Ident, make_ident: Ident) -> Do:
    vim_geometry = yield NS.lift(vim_pane_geometry.value_or_default())
    pane: ViewTree[Layout, Pane] = ViewTree.pane(Pane.cons(ident=ident, open=True))
    vim_layout = ViewTree.layout(Layout.cons(ident, vertical=True, geometry=vim_geometry), sub=List(pane))
    make = ViewTree.pane(Pane.cons(ident=make_ident, pin=True))
    make_layout = ViewTree.layout(Layout.cons(make_ident, vertical=True), sub=List(make))
    layout = ViewTree.layout(Layout.cons(layout_ident, vertical=False), sub=List(vim_layout, make_layout))
    yield NS.modify(__.append1.spaces(Space.cons(ident, List(Window.cons(ident, layout=layout)))))


def no_handler(error: str) -> Prog[None]:
    ribo_log.error(error)
    return Prog.unit


@prog.do(None)
def init_vim_pane(handler: Program) -> Do:
    pid = yield Ribo.lift_nvimio(vim_pid())
    ident = StrIdent('vim')
    layout_ident = StrIdent('root')
    make_ident = StrIdent('make')
    yield Ribo.lift_comp(insert_default_ui(ident, layout_ident, make_ident), UiData)
    yield handler(ident, pid, True)


@prog.do(None)
def run_init() -> Do:
    handler = yield find_handler_e(__.create_vim_pane(), 'insert_vim_pane')
    yield handler.cata(
        no_handler,
        init_vim_pane,
    )


@prog.do(None)
def init() -> Do:
    allow_init = yield Ribo.setting_prog(init_default_ui)
    yield run_init() if allow_init else Prog.unit


__all__ = ('init',)



from typing import Callable, Any, TypeVar

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.state import NS
from ribosome.nvim.api.function import nvim_call_cons_strict
from ribosome.nvim.api.ui import current_cursor
from ribosome.nvim.api.util import cons_decode_str, cons_decode_str_list
from ribosome.nvim.io.api import N
from ribosome.compute.ribosome_api import Ribo

from amino import _, Either, do, Do, Dat
from amino.json import encode_json
from amino.logging import module_log

from myo.components.command.compute.tpe import CommandRibosome
from myo.config.settings import vim_test_filename_modifier

log = module_log()
A = TypeVar('A')
D = TypeVar('D')


def vim_test_call(cons: Callable[[Any], Either[str, A]], fun: str, *args: str) -> NvimIO[str]:
    return nvim_call_cons_strict(cons, f'test#{fun}', *args)


def vim_test_call_wrap(cons: Callable[[Any], Either[str, A]], fun: str, *args: str) -> NvimIO[str]:
    return nvim_call_cons_strict(cons, f'MyoTest{fun}', *args)


class VimTestPosition(Dat['VimTestPosition']):

    def __init__(self, file: str, line: str, col: str) -> None:
        self.file = file
        self.line = line
        self.col = col


@do(NS[CommandRibosome, VimTestPosition])
def vim_test_position() -> Do:
    fn_mod = yield Ribo.setting(vim_test_filename_modifier)
    file = yield NS.lift(nvim_call_cons_strict(cons_decode_str, 'expand', f'%{fn_mod}'))
    line, col = yield NS.lift(current_cursor())
    return VimTestPosition(file, line, col)


@do(NvimIO[None])
def assemble_vim_test_line(position: VimTestPosition) -> Do:
    runner = yield vim_test_call_wrap(cons_decode_str, 'DetermineRunner', position.file)
    exe = yield vim_test_call_wrap(cons_decode_str, 'Executable', runner)
    pos_json = yield N.from_either(encode_json(position))
    pre_args = yield vim_test_call_wrap(cons_decode_str_list, f'BuildPosition', runner, pos_json.native)
    args = yield vim_test_call_wrap(cons_decode_str_list, f'BuildArgs', runner, pre_args)
    return args.cons(exe).join_tokens


@do(NS[CommandRibosome, None])
def vim_test_line() -> Do:
    position = yield vim_test_position()
    yield NS.lift(assemble_vim_test_line(position))


__all__ = ('vim_test_line', 'test_determine_runner', 'test_executable', 'build_position', 'build_args')



from amino import do, Do, Boolean, Path, Just, L, _, Maybe, Nothing
from amino.lenses.lens import lens
from amino.case import Case
from amino.logging import module_log

from chiasma.io.state import TS
from myo.components.tmux.data import TmuxData

from chiasma.pane import pane_by_ident
from chiasma.commands.pane import send_keys, pipe_pane
from chiasma.io.util import tmuxio_repeat_timeout
from chiasma.io.compute import TmuxIO

from ribosome.compute.api import prog
from ribosome.config.component import ComponentData
from ribosome.nvim.io.state import NS

from myo.env import Env
from myo.command.run_task import RunTaskDetails, UiSystemTaskDetails, UiShellTaskDetails, RunTask
from myo.data.command import Command, Pid
from myo.ui.data.view import Pane
from myo.tmux.pid import process_pid

log = module_log()


# FIXME waiting for the pid here is not practical. if the pane was created while executing the command, the shell
# startup time may be very long, while very short commands like `print` may be terminated before the first check is
# done.
@do(TS[ComponentData[Env, TmuxData], Maybe[Pid]])
def run_system_task(cmd: Command, pane: Pane, log_path: Maybe[Path]) -> Do:
    tpane = yield pane_by_ident(pane.ident).zoom(lens.comp.views)
    id = yield TS.from_maybe(tpane.id, f'no tmux id for pane {pane.ident}')
    yield log_path.map(lambda a: TS.lift(pipe_pane(id, a))).get_or_strict(TS.unit)
    yield TS.lift(send_keys(id, cmd.lines))
    yield TS.lift(TmuxIO.flush())
    yield TS.lift(
        tmuxio_repeat_timeout(lambda: process_pid(id), lambda a: a.present, '', timeout=1.)
        .recover_error(lambda a: Nothing)
    )


class run_task_in_pane(Case[RunTaskDetails, Maybe[Pid]], alg=RunTaskDetails):

    def __init__(self, task: RunTask) -> None:
        self.task = task

    @do(TS[ComponentData[Env, TmuxData], Maybe[Pid]])
    def ui_system_task_details(self, details: UiSystemTaskDetails) -> Do:
        yield run_system_task(self.task.command, details.pane, Just(self.task.log))

    @do(TS[ComponentData[Env, TmuxData], Maybe[Pid]])
    def ui_shell_task_details(self, details: UiShellTaskDetails) -> Do:
        yield run_system_task(self.task.command, details.pane, Just(self.task.log))

    @do(TS[ComponentData[Env, TmuxData], Maybe[Pid]])
    def case_default(self, task: RunTaskDetails) -> Do:
        yield TS.error(f'cannot run {task} in tmux')


@prog
@do(NS[ComponentData[Env, TmuxData], Maybe[Pid]])
def tmux_run_command(task: RunTask) -> Do:
    yield run_task_in_pane(task)(task.details).nvim


def tmux_can_run(task: RunTask) -> bool:
    return isinstance(task.details, (UiSystemTaskDetails, UiShellTaskDetails))


__all__ = ('tmux_run_command', 'tmux_can_run')



from typing import Union, Generic, TypeVar, Tuple

from amino import Path, List, Dat, Maybe, Nil, Nothing

A = TypeVar('A')
B = TypeVar('B')


class OutputLine(Generic[A], Dat['OutputLine']):

    def __init__(self, text: str, meta: A, indent: int=0) -> None:
        self.text = text
        self.meta = meta
        self.indent = indent


class Location(Dat['Location']):

    @staticmethod
    def cons(
            path: Path,
            line: int,
            col: int,
    ) -> 'Location':
        return Location(
            path,
            line,
            col,
        )

    def __init__(self, path: Path, line: int, col: int) -> None:
        self.path = path
        self.line = line
        self.col = col

    @property
    def coords(self) -> Tuple[int, int]:
        return self.line, self.col


class OutputEvent(Generic[A, B], Dat['OutputEvent[A, B]']):

    @staticmethod
    def cons(
            meta: B,
            lines: List[OutputLine[A]]=Nil,
            location: Maybe[Location]=Nothing,
    ) -> 'OutputEvent[A]':
        return OutputEvent(meta, lines, location)

    def __init__(self, meta: B, lines: List[OutputLine[A]], location: Maybe[Location]) -> None:
        self.meta = meta
        self.lines = lines
        self.location = location


__all__ = ('OutputLine', 'OutputEvent')



from myo.output.config import LangConfig
from myo.output.lang.python.report import python_report
from myo.output.format.path import project_relative_path
from myo.output.lang.python.syntax import python_syntax
from myo.output.lang.scala.report import scala_report
from myo.output.lang.scala.syntax import scala_syntax
from myo.output.lang.python.first_error import python_first_error
from myo.output.lang.haskell.report import haskell_report
from myo.output.lang.haskell.syntax import haskell_syntax

from amino import List
from amino.util.tpe import qualified_name


python_config = LangConfig.cons(
    'python',
    output_reporter=qualified_name(python_report),
    output_syntax=qualified_name(python_syntax),
    output_first_error=qualified_name(python_first_error),
)

scala_config = LangConfig.cons(
    'scala',
    output_reporter=qualified_name(scala_report),
    output_syntax=qualified_name(scala_syntax),
)

haskell_config = LangConfig.cons(
    'haskell',
    output_reporter=qualified_name(haskell_report),
    output_syntax=qualified_name(haskell_syntax),
)

default_lang_configs = List(
    python_config,
    scala_config,
    haskell_config,
)

global_config_defaults = LangConfig.cons(
    'global',
    output_path_formatter=qualified_name(project_relative_path),
)

__all__ = ('default_lang_configs', 'global_config_defaults',)



from amino import Do

from ribosome.compute.api import prog

from chiasma.util.id import IdentSpec

from myo.util.id import ensure_ident_prog
from myo.components.ui.compute.pane import view_prog


@prog.do(None)
def focus(ident_spec: IdentSpec) -> Do:
    ident = yield ensure_ident_prog(ident_spec)
    handler = yield view_prog(ident, lambda a: a.focus_pane, 'focus_pane')
    yield handler(ident)


__all__ = ('focus',)



# from __future__ import annotations
from typing import Tuple

from amino import do, Do, Dat, Maybe, Just, Nothing

from chiasma.util.id import IdentSpec
from chiasma.open_pane import ui_open_pane
from chiasma.data.view_tree import ViewTree
from amino.logging import module_log
from amino.case import Case

from ribosome.compute.api import prog
from ribosome.compute.ribosome_api import Ribo
from ribosome.nvim.io.state import NS
from ribosome.compute.prog import Prog

from myo.util import Ident
from myo.ui.data.ui_data import UiData
from myo.ui.pane import map_window_trees, find_in_ui
from myo.components.ui.compute.pane import render_view
from myo.ui.data.view_tree import MyoViewTree, MyoPaneNode
from myo.ui.data.view import Pane
from myo.util.id import ensure_ident_prog
from myo.ui.data.window import Window
from myo.components.ui.compute.handler import ui_handler

log = module_log()


class OpenPaneOptions(Dat['OpenPaneOptions']):

    def __init__(self) -> None:
        pass


class match_pane_ident(Case[MyoViewTree, Maybe[Pane]], alg=ViewTree):

    def __init__(self, ident: Ident) -> None:
        self.ident = ident

    def pane(self, pane: MyoPaneNode) -> Maybe[Pane]:
        return Just(pane.data) if pane.data.ident == self.ident else Nothing

    def case_default(self, node: MyoViewTree) -> Maybe[Pane]:
        return Nothing


@do(NS[UiData, Tuple[bool, Window, Pane]])
def chiasma_open_pane(ident: Ident) -> Do:
    pane_m = yield find_in_ui(match_pane_ident(ident))
    (space, window, pane) = yield NS.m(pane_m, lambda: f'no pane named `{ident}`')
    if not pane.open:
        yield map_window_trees(lambda a: ui_open_pane(ident)(a).get_or_strict(a)).nvim
    return (not pane.open), window, pane


@prog.do(bool)
def pane_was_closed(ident: Ident, window: Window, pane: Pane) -> Do:
    pane_open = yield ui_handler(ident, lambda a: a.window_pane_open, 'window_pane_open')
    open = yield pane_open(window, pane)
    return not open


@prog.do(None)
def open_pane(ident_spec: IdentSpec, options: OpenPaneOptions) -> Do:
    ident = yield ensure_ident_prog(ident_spec)
    was_closed, window, pane = yield Ribo.lift_comp(chiasma_open_pane(ident), UiData)
    inconsistent = yield Prog.pure(True) if was_closed else pane_was_closed(ident, window, pane)
    if inconsistent:
        yield render_view(ident)


__all__ = ('open_pane',)



from amino import do, Do, List
from amino.logging import module_log

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.nvim.api.util import cons_decode_str, cons_decode_str_list

from myo.components.command.compute.vim_test import vim_test_call

log = module_log()


@prog
@do(NS[None, str])
def test_determine_runner(file: str) -> Do:
    yield NS.lift(vim_test_call(cons_decode_str, 'determine_runner', file))


@prog
@do(NS[None, str])
def test_executable(runner: str) -> Do:
    yield NS.lift(vim_test_call(cons_decode_str, f'{runner}#executable'))


@prog
@do(NS[None, List[str]])
def test_build_position(runner: str, pos: dict) -> Do:
    yield NS.lift(vim_test_call(cons_decode_str_list, f'{runner}#build_position', 'nearest', pos))


@prog
@do(NS[None, str])
def test_build_args(runner: str, args: list) -> Do:
    yield NS.lift(vim_test_call(cons_decode_str_list, f'{runner}#build_args', args))


__all__ = ('init',)



import os
import tempfile
import subprocess
from typing import TypeVar

from psutil import Process

from chiasma.util.id import IdentSpec

from amino import do, Do, Maybe, __, _, Path, IO, L, List, Boolean, Lists, Nil
from amino.dat import Dat
from amino.case import Case
from amino.logging import module_log
from amino.util.numeric import parse_int

from ribosome.compute.api import prog
from ribosome.compute.program import Program
from ribosome.nvim.io.state import NS
from ribosome.process import Subprocess, SubprocessResult
from ribosome.compute.ribosome_api import Ribo
from ribosome.compute.ribosome import Ribosome
from ribosome.compute.prog import Prog

from myo.util import Ident
from myo.config.handler import find_handler
from myo.data.command import (Command, Interpreter, SystemInterpreter, ShellInterpreter, VimInterpreter, Pid,
                              RunningCommand)
from myo.components.ui.compute.pane import ui_pane_by_ident
from myo.command.run_task import (RunTaskDetails, UiSystemTaskDetails, UiShellTaskDetails, VimTaskDetails,
                                  SystemTaskDetails, is_system_task)
from myo.components.command.compute.history import push_history
from myo.command.run_task import RunTask
from myo.components.command.data import CommandData
from myo.components.ui.compute.open_pane import open_pane, OpenPaneOptions
from myo.config.component import MyoComponent
from myo.env import Env
from myo.components.command.compute.tpe import CommandRibosome
from myo.command.history import history_entry
from myo.util.id import ensure_ident_prog

log = module_log()
D = TypeVar('D')


def internal_can_run(task: RunTask) -> Boolean:
    return Boolean.isinstance(task.details, SystemTaskDetails)


class run_task(Case, alg=RunTaskDetails):

    def __init__(self, task: RunTask) -> None:
        self.task = task

    @do(NS[D, None])
    def system_task_details(self, details: SystemTaskDetails) -> Do:
        def popen(exe: str, args: List[str]) -> IO[SubprocessResult[None]]:
            return Subprocess(Path(exe), args, None, None).execute(stderr=subprocess.STDOUT)
        parts = yield NS.from_either(
            self.task.command.lines
            .map(Lists.tokens)
            .traverse(lambda a: a.detach_head, Maybe)
            .to_either(lambda: f'empty command line in {self.task.command}')
        )
        result = yield NS.from_io(parts.traverse2(popen, IO))
        output = result.flat_map(_.stdout)
        yield NS.from_io(IO.delay(self.task.log.write_text, output.join_lines))
        yield NS.unit

    @do(NS[D, None])
    def case_default(self, details: RunTaskDetails) -> Do:
        yield NS.unit


@prog.unit
@do(NS[None, None])
def run_internal_command(task: RunTask) -> Do:
    yield run_task(task)(task.details)


def process_alive(pid: Pid) -> IO[bool]:
    return IO.delay(Process, pid.value).replace(True).recover(lambda a: False)


@prog
@do(NS[CommandRibosome, bool])
def command_running(cmd: Command) -> Do:
    running_cmd = yield Ribo.inspect_comp(lambda a: a.running.find(lambda b: b.ident == cmd.ident))
    pid = running_cmd.flat_map(lambda a: a.pid)
    running = pid.cata(process_alive, IO.pure(False))
    yield NS.from_io(running)


class RunCommandOptions(Dat['RunCommandOptions']):

    @staticmethod
    def cons(
            interpreter: str=None,
            langs: List[str]=None,
    ) -> 'RunCommandOptions':
        return RunCommandOptions(Maybe.optional(interpreter), Maybe.optional(langs))

    def __init__(
            self,
            interpreter: Maybe[str],
            langs: Maybe[List[str]],
    ) -> None:
        self.interpreter = interpreter
        self.langs = langs


def system_task_details() -> Prog[SystemTaskDetails]:
    return Prog.pure(SystemTaskDetails())


@do(Prog[RunTaskDetails])
def ui_system_task_details(target: Ident) -> Do:
    pane = yield ui_pane_by_ident(target)
    yield open_pane(pane.ident, OpenPaneOptions())
    return UiSystemTaskDetails(pane)


@do(Prog[RunTaskDetails])
def ui_shell_task_details(cmd_ident: Ident, target: Ident) -> Do:
    shell = yield Ribo.lift_comp(NS.inspect_f(__.command_by_ident(target)), CommandData)
    running = yield command_running(shell)
    if not running:
        yield run_command(shell)
    target = yield Prog.from_maybe(shell.interpreter.target,
                                   lambda: f'shell `{shell.ident}` for command `{cmd_ident}` has no pane')
    pane = yield ui_pane_by_ident(target)
    yield open_pane(pane.ident, OpenPaneOptions())
    return UiShellTaskDetails(shell, pane)


class run_task_details(Case[Interpreter, Program[RunTaskDetails]], alg=Interpreter):

    def __init__(self, cmd: Command) -> None:
        self.cmd = cmd

    @do(Prog[RunTaskDetails])
    def system_interpreter(self, interpreter: SystemInterpreter) -> Do:
        yield interpreter.target / ui_system_task_details | (lambda: system_task_details())

    def shell_interpreter(self, interpreter: ShellInterpreter) -> Program[RunTaskDetails]:
        return ui_shell_task_details(self.cmd.ident, interpreter.target)

    @do(Prog[RunTaskDetails])
    def vim_interpreter(self, interpreter: VimInterpreter) -> Do:
        yield Prog.pure(VimTaskDetails())


@do(IO[Path])
def mk_log_file(base: str) -> Do:
    uid = yield IO.delay(os.getuid)
    tmp_dir = yield IO.delay(tempfile.gettempdir)
    tmp_path = Path(tmp_dir) / f'myo-{uid}' / base
    yield IO.delay(tmp_path.mkdir, exist_ok=True, parents=True)
    (fh, fname) = yield IO.delay(tempfile.mkstemp, prefix='pane-', dir=str(tmp_path))
    return Path(fname)


@do(NS[CommandRibosome, Path])
def insert_log(ident: Ident) -> Do:
    base = yield Ribo.inspect_comp(_.uuid)
    path = yield NS.from_io(mk_log_file(str(base)))
    yield Ribo.modify_comp(__.log(ident, path))
    return path


@do(NS[CommandRibosome, Path])
def task_log(ident: Ident) -> Do:
    existing = yield Ribo.inspect_comp(__.log_by_ident(ident))
    yield existing / NS.pure | L(insert_log)(ident)


@do(NS[CommandRibosome, Path])
def store_running_command(cmd: Command, pid: Maybe[Pid], system: bool) -> Do:
    current = yield Ribo.inspect_comp(lambda a: a.running)
    new = current.filter_not(lambda a: a.ident == cmd.ident).cat(RunningCommand(cmd.ident, pid, system))
    yield Ribo.modify_comp(lambda a: a.set.running(new))


@do(Prog[None])
def run_command(cmd: Command) -> Do:
    task_details = yield run_task_details(cmd)(cmd.interpreter)
    cmd_log = yield Ribo.lift(task_log(cmd.ident), CommandData)
    task = RunTask(cmd, cmd_log, task_details)
    handler = yield find_handler(__.run(task), str(task))
    log.debug(f'running `{task}` with `{handler}`')
    yield Ribo.autocmd_prog('MyoRunCommand')
    pid = yield handler(task)
    yield Ribo.lift(store_running_command(cmd, pid, is_system_task(task_details)), CommandData)
    yield push_history(cmd, cmd.interpreter) if cmd.history else Prog.unit
    yield Prog.unit


@prog
def command_by_ident(ident: Ident) -> NS[Ribosome[Env, MyoComponent, CommandData], Command]:
    return Ribo.inspect_comp_e(__.command_by_ident(ident))


@prog.do(None)
def run(ident_spec: IdentSpec, options: RunCommandOptions) -> Do:
    ident = yield ensure_ident_prog(ident_spec)
    cmd = yield command_by_ident(ident)
    yield run_command(cmd)


class RunLineOptions(Dat['RunLineOptions']):

    def __init__(
            self,
            pane: Maybe[Ident],
            shell: Maybe[Ident],
            line: Maybe[str],
            lines: Maybe[List[str]],
            langs: Maybe[List[str]],
            history: Maybe[bool],
    ) -> None:
        self.pane = pane
        self.shell = shell
        self.line = line
        self.lines = lines
        self.langs = langs
        self.history = history


@prog.do(None)
def run_line(options: RunLineOptions) -> Do:
    interpreter = options.shell.cata_f(ShellInterpreter, lambda: SystemInterpreter(options.pane))
    cmd = Command.cons(
        Ident.generate(),
        interpreter,
        lines=options.lines.o(options.line / List) | Nil,
        langs=options.langs | Nil,
        history=options.history | True,
    )
    yield run_command(cmd)


@prog.do(None)
def rerun(index_spec: int=None) -> Do:
    index = yield Prog.e(parse_int(index_spec or 0).lmap(lambda a: f'invalid history index: {a}'))
    entry = yield Ribo.lift_comp(history_entry(index), CommandData)
    yield run_command(entry.cmd)


__all__ = ('run', 'run_line', 'rerun',)



from typing import Any

from lark import Lark, Tree
from lark.lexer import Token

from amino import List, do, Do, Nil, Try, Either, Lists, Right, Left, Maybe
from amino.logging import module_log

from myo.components.command.compute.tpe import CommandRibosome
from myo.output.data.output import OutputEvent, Location
from myo.components.command.compute.parse_handlers import PathFormatter
from myo.output.data.report import DisplayLine, PlainDisplayLine
from myo.output.parser.haskell import HaskellEvent, HaskellLine
from myo.output.lang.haskell.grammar import grammar

from ribosome.nvim.io.state import NS

log = module_log()
parser = Lark(grammar)  # type: ignore


@do(NS[CommandRibosome, List[DisplayLine]])
def format_location(location: Location, path_formatter: PathFormatter) -> Do:
    path = yield path_formatter(location.path)
    return List(PlainDisplayLine(f'{path}  {location.line + 1}'))


def tfilt(tree: Tree, key: str) -> List[Tree]:
    return Lists.wrap(tree.children).filter(lambda c: isinstance(c, Tree) and c.data == key)


def first(trees: List[Tree]) -> List[Tree]:
    return trees.flat_map(lambda a: Lists.wrap(a.children).head)


def token_values(trees: List[Any]) -> List[str]:
    return trees.filter_type(Token).map(lambda a: a.value)


def tnames(tree: Tree) -> List[str]:
    return token_values(first(tfilt(tree, 'name')))


def qnamedata(tree: Tree) -> Maybe[str]:
    return token_values(first(tfilt(tree, 'qnamedata'))).head


def qnames(tree: Tree) -> List[str]:
    return tfilt(tree, 'qname').flat_map(qnamedata)


def tlift(tree: Tree, key: str) -> Either[str, Tree]:
    return tfilt(tree, key).head.to_either(f'no child `key`')


@do(Either[str, List[Tree]])
def foundreq(tree: Tree) -> Do:
    req, found = yield qnames(tree).lift_all(0, 1).to_either('invalid name count')
    return List(
        'type mismatch',
        found.replace("\n", ' '),
        req.replace("\n", ' '),
    )


@do(Either[str, List[Tree]])
def notypeclass(tree: Tree) -> Do:
    pt = yield tlift(tree, 'parenstype')
    names = tnames(pt)
    tc, tpe = yield names.lift_all(0, 1).to_either('too few names')
    trigger = yield qnames(tree).head.to_either('no trigger')
    return List(
        f'!instance: {trigger}',
        f'{tc} {tpe}'
    )


@do(Either[str, List[str]])
def from_grammar(lines: List[str]) -> Do:
    tree = yield Try(parser.parse, lines.join_lines)
    spec = yield tlift(tree, 'specific')
    candidate = yield Lists.wrap(spec.children).head.to_either('empty specific')
    dotted = yield (
        Lists.wrap(candidate.children).lift(1).to_either('empty dotted')
        if candidate.data == 'dotted' else
        Left('candidate is not dotted')
    )
    tpe = dotted.data
    yield (
        foundreq(dotted)
        if tpe == 'foundreq' else
        notypeclass(dotted)
        if tpe == 'notypeclass' else
        Right(Nil)
    )


def format_info(lines: List[str]) -> List[str]:
    return from_grammar(lines).get_or_strict(lines)


@do(NS[CommandRibosome, List[DisplayLine]])
def haskell_report(path_formatter: PathFormatter, output: OutputEvent[HaskellLine, HaskellEvent]) -> Do:
    info = format_info(output.meta.info.map(lambda a: a.meta.message)).indent(2).map(PlainDisplayLine)
    location = yield output.location.map(lambda a: format_location(a, path_formatter)).get_or(NS.pure, Nil)
    return location + info


__all__ = ('haskell_report',)



from amino import do, Do
from amino.state import EitherState
from amino.lenses.lens import lens
from amino.boolean import false

from chiasma.util.id import Ident, IdentSpec, ensure_ident_or_generate

from ribosome.compute.api import prog
from ribosome.config.component import ComponentData
from ribosome.nvim.io.state import NS

from myo.ui.data.ui_data import UiData
from myo.ui.data.window import Window
from myo.ui.pane import ui_modify_pane
from myo.env import Env
from myo.components.ui.compute.pane import render_view


def ui_close_pane(ident: Ident) -> EitherState[str, UiData, Window]:
    return ui_modify_pane(ident, lens.open.set(false))


@prog
def ui_close_pane_prog(ident_spec: IdentSpec) -> NS[ComponentData[Env, UiData], Window]:
    ident = ensure_ident_or_generate(ident_spec)
    return ui_close_pane(ident).nvim.zoom(lens.comp)


@prog.do(None)
def close_pane(ident_spec: IdentSpec) -> Do:
    ident = ensure_ident_or_generate(ident_spec)
    yield ui_close_pane_prog(ident)
    yield render_view(ident)


__all__ = ('close_pane',)



from typing import Callable

from chiasma.ui.view import UiPane, UiLayout, UiView
from chiasma.ui.state import ViewState
from chiasma.ui.view_geometry import ViewGeometry
from chiasma.util.id import ensure_ident_or_generate, IdentSpec

from amino import Boolean, ADT, Maybe, Path
from amino.tc.base import tc_prop

from myo.util import Ident


class View(ADT['View']):

    def __init__(self, ident: Ident, state: ViewState, geometry: ViewGeometry, ui: str) -> None:
        self.ident = ident
        self.state = state
        self.geometry = geometry
        self.ui = ui


class Layout(View):

    @staticmethod
    def cons(
            ident: IdentSpec=None,
            state: ViewState=None,
            geometry: ViewGeometry=None,
            ui: str='tmux',
            vertical: bool=True,
    ) -> 'Layout':
        return Layout(
            ensure_ident_or_generate(ident),
            state or ViewState.cons(),
            geometry or ViewGeometry.cons(),
            ui,
            Boolean(vertical),
        )

    def __init__(
            self,
            ident: Ident,
            state: ViewState,
            geometry: ViewGeometry,
            ui: str,
            vertical: Boolean,
    ) -> None:
        super().__init__(ident, state, geometry, ui)
        self.vertical = vertical

    @property
    def vertical_str(self) -> str:
        return 'vertical' if self.vertical else 'horizontal'


class Pane(View):

    @staticmethod
    def cons(
            ident: IdentSpec=None,
            state: ViewState=None,
            geometry: ViewGeometry=None,
            ui: str='tmux',
            open: bool=False,
            cwd: Path=None,
            pin: bool=False,
    ) -> 'Pane':
        return Pane(
            ensure_ident_or_generate(ident),
            state or ViewState.cons(),
            geometry or ViewGeometry.cons(),
            ui,
            Boolean(open),
            Maybe.optional(cwd),
            Boolean(pin),
        )

    def __init__(
            self,
            ident: Ident,
            state: ViewState,
            geometry: ViewGeometry,
            ui: str,
            open: Boolean,
            cwd: Maybe[Path],
            pin: Boolean,
    ) -> None:
        super().__init__(ident, state, geometry, ui)
        self.open = open
        self.cwd = cwd
        self.pin = pin


class MyoUiPane(UiPane, tpe=Pane):

    @tc_prop
    def ident(self, a: Pane) -> Ident:
        return a.ident

    @tc_prop
    def open(self, a: Pane) -> Ident:
        return a.open

    def cwd(self, a: Pane) -> Maybe[Path]:
        return a.cwd

    def pin(self, a: Pane) -> bool:
        return a.pin

    def set_open(self, a: Pane, state: Boolean) -> Pane:
        return a.set.open(state)


class MyoUiLayout(UiLayout, tpe=Layout):
    pass


class MyoPaneUiView(UiView, tpe=Pane):

    def state(self, a: Pane) -> ViewState:
        return a.state

    def geometry(self, a: Pane) -> ViewGeometry:
        return a.geometry

    def ident(self, a: Pane) -> Ident:
        return a.ident


class MyoLayoutUiView(UiView, tpe=Layout):

    def state(self, a: Layout) -> ViewState:
        return a.state

    def geometry(self, a: Layout) -> ViewGeometry:
        return a.geometry

    def ident(self, a: Layout) -> Ident:
        return a.ident


def has_ident(ident_spec: IdentSpec) -> Callable[[View], bool]:
    ident = ensure_ident_or_generate(ident_spec)
    def has_ident(view: View) -> bool:
        return view.ident == ident
    return has_ident


__all__ = ('View', 'Layout', 'Pane', 'ViewGeometry', 'has_ident',)



from typing import Tuple, Callable

from amino import do, Do, curried, Maybe, Just, Nothing, _

from chiasma.data.view_tree import ViewTree
from amino.state import State
from amino.lenses.lens import lens
from amino.util.tuple import lift_tuple

from ribosome.compute.api import prog
from ribosome.compute.program import Program
from ribosome.config.component import ComponentData
from ribosome.data.plugin_state import PluginState
from ribosome.compute.ribosome_api import Ribo
from ribosome.nvim.io.state import NS
from ribosome.compute.prog import Prog

from myo.util import Ident
from myo.ui.data.window import Window
from myo.ui.data.space import Space
from myo.ui.data.view import Pane, Layout, has_ident
from myo.ui.pane import find_in_ui
from myo.env import Env
from myo.config.component import MyoComponent
from myo.ui.data.ui_data import UiData
from myo.ui.data.view_tree import MyoViewTree
from myo.ui.ui import Ui


@curried
def match_ident(ident: Ident, node: MyoViewTree) -> Maybe[MyoViewTree]:
    return Just(node) if has_ident(ident)(node.data) else Nothing


@do(State[UiData, Maybe[Tuple[Space, Window, ViewTree[Layout, Pane]]]])
def view_path_by_ident(ident: Ident) -> Do:
    yield find_in_ui(match_ident(ident))


@prog
@do(State[PluginState[Env, MyoComponent], Program])
def config_uis() -> Do:
    components = yield State.inspect(_.components.all)
    yield State.pure(components // _.config // _.ui)


@prog.do(Tuple[Space, Window, Ui])
def view_owners(ident: Ident) -> Do:
    view_path_m = yield Ribo.lift_comp(view_path_by_ident(ident).nvim, UiData)
    space, window, view = yield Prog.from_maybe(view_path_m, f'no view with ident `{ident}`')
    uis = yield config_uis()
    owns_handlers = uis / _.owns_view
    owns = yield owns_handlers.traverse(lambda a: a(view.data), Prog)
    owners = uis.zip(owns).filter2(lambda a, b: b).flat_map(lift_tuple(0))
    owner = yield (
        Prog.from_maybe(owners.head, f'no owner for view {ident}')
        if owners.length <= 1
        else Prog.error('multiple owners for view {ident}')
    )
    yield Prog.pure((space, window, owner))


@prog.do(Program)
def view_prog(ident: Ident, prog: Callable[[Ui], Maybe[Program]], desc: str) -> Do:
    space, window, owner = yield view_owners(ident)
    yield Prog.m(prog(owner), lambda: '{type(owner)} has no handler for `{desc}`')


@prog.do(None)
def render_view(ident: Ident) -> Do:
    space, window, owner = yield view_owners(ident)
    renderer = yield Prog.from_maybe(owner.render, f'no renderer for {owner}')
    yield renderer(space.ident, window.ident, window.layout)


@prog
@do(NS[ComponentData[Env, UiData], Pane])
def ui_pane_by_ident(ident: Ident) -> Do:
    result = yield view_path_by_ident(ident).nvim.zoom(lens.comp)
    pane = (
        result
        .to_either(f'no pane for `{ident}`')
        .flat_map(lift_tuple(2))
        .map(_.data)
        .to_either(f'invalid result from `pane_path_by_ident`')
    )
    yield NS.e(pane)


__all__ = ('config_uis', 'pane_owners', 'render_pane', 'view_prog',)



from amino import Dat, Maybe

from chiasma.util.id import Ident
from chiasma.data.tmux import Views


class TmuxData(Dat['TmuxData']):

    @staticmethod
    def cons(views: Views=None, vim_pane: Ident=None) -> 'TmuxData':
        return TmuxData(
            views or Views.cons(),
            Maybe.optional(vim_pane)
        )

    def __init__(self, views: Views, vim_pane: Maybe[Ident]) -> None:
        self.views = views
        self.vim_pane = vim_pane


__all__ = ('TmuxData',)



from amino import do, Do, List, _, Nil, Just

from ribosome.nvim.scratch import show_in_scratch_buffer_default
from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.compute.prog import Prog
from ribosome.compute.ribosome_api import Ribo
from ribosome.compute.program import bind_programs

from myo.config.handler import find_handlers
from myo.data.info import InfoWidget
from myo.env import Env


@do(NS[Env, None])
def display_info(widgets: List[InfoWidget]) -> Do:
    yield NS.lift(show_in_scratch_buffer_default(widgets // _.lines, Just(.5)))


@prog.do(None)
def collect_info() -> Do:
    programs = yield find_handlers(_.info)
    yield bind_programs(programs, Nil)


@prog.do(None)
def info() -> Do:
    widgets = yield collect_info()
    yield Ribo.trivial(display_info(widgets))


__all__ = ('info',)



from uuid import uuid4, UUID

from amino import Dat, Nil, Map, Path, List, Either, _, Maybe
from amino.list import replace_one

from myo.data.command import Command, HistoryEntry, RunningCommand, CommandConfig
from myo.output.data.report import ParseReport
from myo.output.config import LangConfig

from chiasma.util.id import Ident

from ribosome.nvim.scratch import ScratchBuffer
from ribosome.nvim.io.state import NS


class OutputData(Dat['OutputData']):

    @staticmethod
    def cons(
            report: ParseReport=None,
            scratch: ScratchBuffer=None,
            current: int=0,
    ) -> 'OutputData':
        return OutputData(
            Maybe.optional(report),
            Maybe.optional(scratch),
            current,
        )

    def __init__(
            self,
            report: Maybe[ParseReport],
            scratch: Maybe[ScratchBuffer],
            current: int,
    ) -> None:
        self.report = report
        self.scratch = scratch
        self.current = current


class CommandData(Dat['CommandData']):

    @staticmethod
    def cons(
            commands: List[Command]=Nil,
            command_configs: List[CommandConfig]=Nil,
            lang_configs: List[LangConfig]=Nil,
            logs: Map[Ident, Path]=Map(),
            uuid: UUID=None,
            history: List[HistoryEntry]=Nil,
            output: OutputData=None,
            running: List[RunningCommand]=Nil,
    ) -> 'CommandData':
        return CommandData(
            commands,
            command_configs,
            lang_configs,
            logs,
            uuid or uuid4(),
            history,
            output or OutputData.cons(),
            running,
        )

    def __init__(
            self,
            commands: List[Command],
            command_configs: List[CommandConfig],
            lang_configs: List[LangConfig],
            logs: Map[Ident, Path],
            uuid: UUID,
            history: List[HistoryEntry],
            output: OutputData,
            running: List[RunningCommand]=Nil,
    ) -> None:
        self.commands = commands
        self.command_configs = command_configs
        self.lang_configs = lang_configs
        self.logs = logs
        self.uuid = uuid
        self.history = history
        self.output = output
        self.running = running

    def command_by_ident(self, ident: Ident) -> Either[str, Command]:
        return self.commands.find(_.ident == ident).to_either(f'no command `{ident}`')

    def replace_command(self, cmd: Command) -> 'CommandData':
        return self.mod.commands(lambda cs: replace_one(cs, _.ident == cmd.ident, cmd)._value() | (lambda: cs.cat(cmd)))

    def log_by_ident(self, ident: Ident) -> Either[str, Path]:
        return self.logs.lift(ident).to_either(f'no log for `{ident}`')

    def log(self, ident: Ident, path: Path) -> 'CommandData':
        return self.append.logs((ident, path))


def running_command(ident: Ident) -> NS[CommandData, Command]:
    return NS.inspect(lambda a: a.running.find(lambda b: b.ident == ident).to_either(f'no running command `{ident}`'))


__all__ = ('CommandData',)



from typing import Callable, TypeVar

from amino import List, Either

from ribosome.nvim.io.state import NS

A = TypeVar('A')
D = TypeVar('D')


def resolve_callbacks(paths: List[str]) -> Either[str, List[Callable[..., NS[D, A]]]]:
    return paths.traverse(Either.import_path, Either)


__all__ = ('resolve_callbacks',)



from amino import do, Do, _, Nil

from ribosome.compute.prog import Prog
from ribosome.compute.api import prog
from ribosome.compute.program import bind_programs

from myo.config.handler import find_handlers


@prog.do(None)
def vim_leave() -> Do:
    programs = yield find_handlers(_.quit)
    yield bind_programs(programs, Nil)


__all__ = ('vim_leave',)



from amino import Dat, List, _, Nil

from myo.ui.data.space import Space


class UiData(Dat['UiData']):

    @staticmethod
    def cons(
        spaces: List[Space]=Nil,
    ) -> 'UiData':
        return UiData(
            spaces,
        )

    def __init__(self, spaces: List[Space]) -> None:
        self.spaces = spaces

    def replace_space(self, space: Space) -> 'UiData':
        return self.copy(spaces=self.spaces.replace_where(space, _.ident == space.ident))


__all__ = ('UiData',)



from chiasma.data.view_tree import ViewTree, LayoutNode, PaneNode

from myo.ui.data.view import Layout, Pane


MyoViewTree = ViewTree[Layout, Pane]
MyoLayoutNode = LayoutNode[Layout, Pane]
MyoPaneNode = PaneNode[Layout, Pane]

__all__ = ('MyoViewTree', 'MyoLayoutNode', 'MyoPaneNode',)



from typing import Callable

from amino import Maybe, do, Do, List, _, Either, Left, Right
from amino.logging import module_log

from myo.config.plugin_state import MyoState
from myo.config.component import MyoComponent

from ribosome.compute.program import Program
from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.nvim.io.api import N
from ribosome.compute.prog import Prog

log = module_log()


@prog
@do(NS[MyoState, List[Program]])
def find_handlers(pred: Callable[[MyoComponent], Maybe[Program]]) -> Do:
    components = yield NS.inspect(_.components.all)
    return components.flat_map(_.config).flat_map(pred)


@prog
@do(NS[MyoState, Either[str, Program]])
def select_handler(eligible: List[Program], desc: str) -> Do:
    def select(h: Program, t: List[Program]) -> Either[str, Program]:
        return (
            Right(h)
            if t.empty else
            Left(f'multiple handlers for `{desc}`: {eligible}')
        )
    yield NS.pure(eligible.detach_head.to_either(f'no handler for `{desc}`').flat_map2(select))


@prog.do(Either[str, Program])
def find_handler_e(pred: Callable[[MyoComponent], Maybe[Program]], desc: str) -> Do:
    eligible = yield find_handlers(pred)
    yield select_handler(eligible, desc)


@prog.do(Program)
def find_handler(pred: Callable[[MyoComponent], Maybe[Program]], desc: str) -> Do:
    handler_e = yield find_handler_e(pred, desc)
    yield Prog.e(handler_e)


__all__ = ('find_handler', 'find_handlers', 'find_handler_e',)



from amino import do, Do

from ribosome.compute.api import prog

from chiasma.util.id import IdentSpec
from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo

from myo.components.command.compute.tpe import CommandRibosome
from myo.util.id import ensure_ident_ns
from myo.components.command.data import running_command
from myo.util.process import kill_process


@prog
@do(NS[CommandRibosome, None])
def kill(ident_spec: IdentSpec) -> Do:
    ident = yield ensure_ident_ns(ident_spec)
    cmd_e = yield Ribo.zoom_comp(running_command(ident))
    cmd = yield NS.e(cmd_e)
    pid = yield NS.m(cmd.pid, lambda: f'cannot kill command `{ident}`: no associated pid')
    yield NS.from_io(kill_process(pid))


__all__ = ('kill',)



grammar = '''
start: (specific | generic)+
specific.5: dotted
generic.1: any+
dotted: dot (foundreq | notypeclass | genericdot)

foundreq.5: "Couldn't match expected type" qname "with actual type" qname
notypeclass.5: "No instance for" parenstype "arising from" words qname
genericdot.1: any+
dot: "•"
qname: sql qnamedata sqr
qnamedata: /[^’]+/
parenstype: "(" name name ")"
name: WORD
words: WORD+
sql: "‘"
sqr: "’"
any: (/[^•]+/ | NEWLINE)
%import common.NEWLINE
%import common.WS
%import common.WORD
%ignore WS
%ignore NEWLINE
'''

__all__ = ('grammar',)



from ribosome.compute.ribosome import Ribosome

from myo.components.tmux.data import TmuxData

from myo.env import Env
from myo.config.component import MyoComponent

TmuxRibosome = Ribosome[Env, MyoComponent, TmuxData]


__all__ = ('TmuxRibosome',)



from amino import List, Map, Maybe, Just, Nothing
from amino.boolean import true

from ribosome.config.component import Component
from ribosome.rpc.api import rpc
from ribosome.compute.program import Program
from ribosome.data.mapping import Mappings

from myo.components.command.compute.add import add_system_command, add_vim_command, add_shell_command
from myo.components.command.compute.run import run, run_line, internal_can_run, run_internal_command, rerun
from myo.config.component import MyoComponent
from myo.components.command.data import CommandData
from myo.components.command.compute.parse import parse
from myo.components.command.compute.output import (current_event_jump, jump_mapping, quit_mapping, quit_output,
                                                 prev_mapping, prev_event, next_mapping, next_event)
from myo.command.run_task import RunTask
from myo.components.command.compute.test import vim_test
from myo.components.command.compute.vim_test_wrappers import (test_determine_runner, test_executable,
                                                              test_build_position, test_build_args)
from myo.components.command.compute.kill import kill
from myo.components.command.compute.reboot import reboot
from myo.components.command.compute.history import history
from myo.components.command.compute.init import command_init
from myo.components.command.compute.history_menu import history_menu


def run_handler_for(task: RunTask) -> Maybe[Program]:
    return Just(run_internal_command) if internal_can_run(task) else Nothing


command: Component[CommandData, MyoComponent] = Component.cons(
    'command',
    state_type=CommandData,
    rpc=List(
        rpc.write(add_vim_command).conf(json=true),
        rpc.write(add_system_command).conf(json=true),
        rpc.write(add_shell_command).conf(json=true),
        rpc.write(run).conf(json=true),
        rpc.write(run_line).conf(name=Just('line'), json=true),
        rpc.write(rerun),
        rpc.write(parse).conf(json=true),
        rpc.write(current_event_jump),
        rpc.write(quit_output),
        rpc.write(prev_event),
        rpc.write(next_event),
        rpc.write(vim_test).conf(json=true),
        rpc.write(kill),
        rpc.write(reboot),
        rpc.write(history_menu),
        rpc.read(history),
        rpc.read(test_determine_runner),
        rpc.read(test_executable),
        rpc.read(test_build_position),
        rpc.read(test_build_args),
    ),
    config=MyoComponent.cons(run=run_handler_for, init=command_init),
    mappings=Mappings.cons(
        (jump_mapping, current_event_jump),
        (quit_mapping, quit_output),
        (prev_mapping, prev_event),
        (next_mapping, next_event),
    ),
)

__all__ = ('command',)



from types import ModuleType
from typing import TypeVar

from amino import List, Either, do, Do
from amino.logging import module_log
from amino.mod import instance_from_module

from myo.output.data.output import OutputEvent
from myo.output.parser.base import Parser, parse_events
from myo.output.config import ParseConfig

log = module_log()
parsers_module = 'myo.output.parser'
A = TypeVar('A')
B = TypeVar('B')


def resolve_parsers(langs: List[str]) -> List[Parser]:
    def imp(mod: str) -> List[ModuleType]:
        return Either.import_module(f'{parsers_module}.{mod}').or_else_call(Either.import_module, mod).to_list
    modules = langs.flat_map(imp)
    return modules.traverse(lambda a: instance_from_module(a, Parser), Either)


@do(Either[str, List[OutputEvent[A, B]]])
def parsers_events(parsers: List[Parser[A, B]], output: List[str]) -> Do:
    yield parsers.flat_traverse(lambda a: parse_events(a, output), Either)


@do(Either[str, List[OutputEvent[A, B]]])
def parse_with_langs(output: List[str], langs: List[str]) -> Do:
    parsers = yield resolve_parsers(langs)
    yield parsers_events(parsers, output)


__all__ = ('parsers_events', 'parse_with_langs', 'ParseConfig',)



from amino import do, Do
from amino.lenses.lens import lens

from ribosome.compute.api import prog
from ribosome.config.component import ComponentData
from ribosome.nvim.io.state import NS

from chiasma.render import render
from myo.components.tmux.data import TmuxData
from chiasma.util.id import Ident
from chiasma.data.view_tree import ViewTree

from myo.ui.data.view import Pane, Layout
from myo.env import Env


@prog
@do(NS[ComponentData[Env, TmuxData], None])
def tmux_render(session: Ident, window: Ident, layout: ViewTree[Pane, Layout]) -> Do:
    yield render(P=Pane, L=Layout)(session, window, layout).zoom(lens.comp.views).nvim


__all__ = ('tmux_render',)



from uuid import uuid4

from amino import Dat, List, Nil, _

from chiasma.util.id import IdentSpec, ensure_ident_or_generate

from myo.ui.data.window import Window
from myo.util import Ident


class Space(Dat['Space']):

    @staticmethod
    def cons(
            ident: IdentSpec=None,
            windows: List[Window]=Nil,
    ) -> 'Space':
        return Space(
            ensure_ident_or_generate(ident),
            windows,
        )

    def __init__(self, ident: Ident, windows: List[Window]) -> None:
        self.ident = ident
        self.windows = windows

    def replace_window(self, window: Window) -> 'Space':
        return self.copy(windows=self.windows.replace_where(window, _.ident == window.ident))


__all__ = ('Space',)



from amino import do, Do

from ribosome.nvim.io.state import NS

from myo.components.command.data import CommandData
from myo.data.command import HistoryEntry, Command


def history_entry(index: int) -> NS[CommandData, HistoryEntry]:
    return NS.inspect_maybe(
        lambda s: s.history.lift(index),
        lambda: f'no history entry for index `{index}`'
    )


@do(NS[CommandData, Command])
def most_recent_command() -> Do:
    entry = yield history_entry(0)
    return entry.cmd


__all__ = ('history_entry', 'most_recent_command',)



from typing import TypeVar

from amino import do, Do, __, List, Nil, Nothing, Just, Either, Right
from amino.logging import module_log

from chiasma.util.id import StrIdent, IdentSpec, ensure_ident

from ribosome.nvim.io.state import NS
from ribosome.compute.api import prog
from ribosome.compute.ribosome_api import Ribo

from myo.components.command.data import CommandData
from myo.components.command.compute.run import run, RunCommandOptions
from myo.data.command import Command, SystemInterpreter, ShellInterpreter
from myo.components.command.compute.vim_test import vim_test_line
from myo.components.command.compute.tpe import CommandRibosome
from myo.config.settings import test_pane, test_ui, test_langs, test_shell

log = module_log()
D = TypeVar('D')
test_ident = StrIdent('<test>')


@do(Either[str, Command])
def cons_test_pane_command(ui: str, target: IdentSpec) -> Do:
    effective_target = yield Right(Nothing) if ui == 'internal' else ensure_ident(target).map(Just)
    return Command.cons(ident=test_ident, interpreter=SystemInterpreter(effective_target))


@do(Either[str, Command])
def cons_test_shell_command(ui: str, shell: IdentSpec) -> Do:
    ident = yield ensure_ident(shell)
    return Command.cons(ident=test_ident, interpreter=ShellInterpreter(ident))


def cons_test_command(ui: str, shell: Either[str, IdentSpec], target: IdentSpec) -> Command:
    return shell.cata(
        lambda e: cons_test_pane_command(ui, target),
        lambda a: cons_test_shell_command(ui, a),
    )


@do(NS[CommandData, None])
def update_test_command(lines: List[str], langs: List[str]) -> Do:
    ui = yield NS.lift(test_ui.value_or_default())
    shell = yield NS.lift(test_shell.value)
    target = yield NS.lift(test_pane.value_or_default())
    existing = yield NS.inspect(__.command_by_ident(test_ident))
    cmd = yield existing.cata(
        lambda e: NS.e(cons_test_command(ui, shell, target)),
        NS.pure,
    )
    global_langs = yield NS.lift(test_langs.value_or_default())
    updated_cmd = cmd.set.lines(lines).set.langs(global_langs + langs)
    log.debug(f'computed test command `{updated_cmd}`')
    yield NS.modify(__.replace_command(updated_cmd))


@do(NS[CommandRibosome, List[str]])
def vim_test_lines() -> Do:
    line = yield vim_test_line()
    return List(line)


@prog
@do(NS[CommandRibosome, None])
def vim_test_command(langs: List[str]) -> Do:
    lines = yield vim_test_lines()
    yield Ribo.zoom_comp(update_test_command(lines, langs))


@prog.do(None)
def vim_test(run_options: RunCommandOptions) -> Do:
    yield vim_test_command(run_options.langs.get_or_strict(Nil))
    yield run(test_ident, run_options)


__all__ = ('vim_test',)



from amino import List, Nil

from ribosome.nvim.io.state import NS
from ribosome.nvim.syntax.syntax import Syntax
from ribosome.nvim.syntax.expr import SyntaxMatch, HiLink, SyntaxLiteral

from myo.components.command.compute.tpe import CommandRibosome

python_include = SyntaxLiteral('syntax include @python syntax/python.vim')
location = SyntaxMatch.cons('MyoLocation', '^.*.*$', 'skipwhite', 'skipnl',
                            contains='MyoPath,MyoLineNumber,MyoFunction', nextgroup='MyoCode')
path = SyntaxMatch.cons('MyoPath', '^.*\ze\( .*$\)\@=', 'contained', containedin='MyoLocation')
line_number = SyntaxMatch.cons('MyoLineNumber', '\( \)\@<=\zs\d\+\ze', 'contained', containedin='MyoLocation')
function = SyntaxMatch.cons('MyoFunction', '\( \d\+ \)\@<=\zs.*$', 'contained', containedin='MyoLocation')
error = SyntaxMatch.cons('MyoError', '^\w\+: .*$', contains='MyoException')
code_line = SyntaxMatch.cons('MyoCode', '^    .*$', 'contained', contains='@python')
exception = SyntaxMatch.cons('MyoException', '^\w\+:', 'contained', containedin='MyoError')

hl_error = HiLink('MyoException', 'Error')
hl_path = HiLink('MyoPath', 'Directory')
hl_line_number = HiLink('MyoLineNumber', 'Directory')
hl_function = HiLink('MyoFunction', 'Type')

syntax = Syntax(
    List(location, path, line_number, function, error, code_line, exception, python_include),
    Nil,
    List(hl_error, hl_path, hl_line_number, hl_function),
)


def python_syntax() -> NS[CommandRibosome, Syntax]:
    return NS.pure(syntax)


__all__ = ('python_syntax',)



from numbers import Number

from amino import do, Do, List, Nil, _, Maybe
from amino.case import Case

from chiasma.data.view_tree import ViewTree, LayoutNode, PaneNode, SubUiNode
from chiasma.ui.view_geometry import ViewGeometry

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.config.component import ComponentData

from myo.env import Env
from myo.ui.data.ui_data import UiData
from myo.data.info import InfoWidget
from myo.ui.data.space import Space
from myo.ui.data.window import Window
from myo.ui.data.view import Layout, Pane


def format_geometry(geo: ViewGeometry) -> str:
    def field(key: str, value: Maybe[Number]) -> Maybe[str]:
        return value.map(lambda v: f'⚬{key} {v}')
    fields = geo._dat__items.flat_map2(field)
    return fields.join_tokens


class format_layout(Case, alg=ViewTree):

    def layout_node(self, node: LayoutNode[Layout, Pane]) -> List[str]:
        direction = '▤' if node.data.vertical else '▥'
        desc = f'{direction} {node.data.ident.str}'
        return (node.sub // self).indent(1).cons(desc)

    def pane_node(self, node: PaneNode[Layout, Pane]) -> List[str]:
        return List(f'◳ {node.data.ident.str} {format_geometry(node.data.geometry)}')

    def sub_ui_node(self, node: SubUiNode[Layout, Pane]) -> List[str]:
        return Nil


def format_window(window: Window) -> List[str]:
    desc = f'□ {window.ident.str}'
    return format_layout.match(window.layout).indent(1).cons(desc)


def format_space(space: Space) -> List[str]:
    desc = f'⬚ {space.ident.str}'
    return (space.windows // format_window).indent(1).cons(desc)


@prog
@do(NS[ComponentData[Env, UiData], InfoWidget])
def ui_info() -> Do:
    spaces = yield NS.inspect(_.comp.spaces)
    tree = spaces // format_space
    yield NS.pure(InfoWidget(List('Ui:') + tree.indent(1)))


__all__ = ('ui_info',)



from chiasma.io.state import TS
from chiasma.util.id import Ident
from chiasma.commands.pane import select_pane

from amino import do, Do

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS

from myo.components.tmux.tpe import TmuxRibosome
from myo.components.tmux.pane import pane_id_by_ident


@prog
@do(NS[TmuxRibosome, None])
def tmux_focus_pane(ident: Ident) -> Do:
    id = yield pane_id_by_ident(ident)
    yield TS.lift(select_pane(id)).nvim


__all__ = ('tmux_focus_pane',)



from amino import Dat, List


class InfoWidget(Dat['InfoWidget']):

    @staticmethod
    def cons(
            lines: List[str],
    ) -> 'InfoWidget':
        return InfoWidget(
            lines,
        )

    def __init__(self, lines: List[str]) -> None:
        self.lines = lines


__all__ = ('InfoWidget',)



from amino import do, Do
from amino.lenses.lens import lens

from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo

from chiasma.util.id import Ident
from chiasma.pane import pane_by_ident

from myo.components.tmux.tpe import TmuxRibosome


@do(NS[TmuxRibosome, int])
def pane_id_by_ident(ident: Ident) -> Do:
    pane = yield Ribo.zoom_comp(pane_by_ident(ident).nvim.zoom(lens.views))
    yield NS.m(pane.id, lambda: f'pane `{ident}` has no id')


__all__ = ('pane_id_by_ident',)



from amino import List, Nil

from ribosome.nvim.io.state import NS
from ribosome.nvim.syntax.syntax import Syntax
from ribosome.nvim.syntax.expr import SyntaxMatch, HiLink, SyntaxLiteral, SyntaxRegion, Highlight

from myo.components.command.compute.tpe import CommandRibosome
from myo.output.lang.scala.report import col_marker, found_marker, separator_marker, foundreq_separator, req_marker


def conceal(name: str, pattern: str, *a: str, **kw: str) -> SyntaxMatch:
    return SyntaxMatch.cons(
        name,
        pattern,
        'conceal',
        'contained',
        *a,
        **kw,
    )

error_end = f'\ze.*\(\|{col_marker}\)'
scala_include = SyntaxLiteral('syntax include @scala syntax/scala.vim')
location = SyntaxMatch.cons(
    'MyoLocation',
    '^.*.*$',
    'skipwhite',
    'skipnl',
    contains='MyoPath,MyoLineNumber',
    nextgroup='MyoError',
)
path = SyntaxMatch.cons('MyoPath', '^.*\ze\( .*$\)\@=', 'contained', containedin='MyoLocation')
line_number = SyntaxMatch.cons('MyoLineNumber', '\( \)\@<=\zs\d\+\ze', 'contained', containedin='MyoLocation')
error = SyntaxRegion.cons(
    'MyoError',
    '^.',
    error_end,
    'contained',
    'skipwhite',
    'skipnl',
    nextgroup='MyoCode',
    contains='MyoSplain,MyoSplainFoundReq',
)
code_line = SyntaxMatch.cons('MyoCode', '.*', 'contained', contains='@scala,MyoColMarker')
col_marker_conceal = conceal('MyoColMarker', col_marker, nextgroup='MyoCol')
col = SyntaxMatch.cons('MyoCol', '.', 'contained')
splain = SyntaxRegion.cons(
    'MyoSplain',
    '\s*!I',
    error_end,
    'contained',
    'skipwhite',
    'skipnl',
    nextgroup='MyoCode',
    contains='MyoSplainParam,MyoSplainCandidate',
)
splain_param = SyntaxMatch.cons(
    'MyoSplainParam',
    '^\s*!I.*',
    'contained',
    containedin='MyoSplain',
    contains='MyoSplainParamMarker',
)
splain_param_marker = SyntaxMatch.cons(
    'MyoSplainParamMarker',
    '!I',
    'contained',
    nextgroup='MyoSplainParamName',
    contains='MyoSplainParamMarkerBang',
)
splain_param_marker_bang = SyntaxMatch.cons(
    'MyoSplainParamMarkerBang',
    '!',
    'contained',
    nextgroup='MyoSplainParamMarkerI',
)
splain_param_marker_i = SyntaxMatch.cons(
    'MyoSplainParamMarkerI',
    'I',
    'contained',
    'skipwhite',
    nextgroup='MyoSplainParamName',
)
splain_param_name = SyntaxMatch.cons(
    'MyoSplainParamName',
    '[^:]\+\ze:',
    'contained',
    nextgroup='MyoSplainParamType',
)
splain_param_type = SyntaxRegion.cons(
    'MyoSplainParamType',
    '.',
    f'\ze.*\(invalid because\|{col_marker}\)',
    'contained',
)
splain_candidate = SyntaxMatch.cons(
    'MyoSplainCandidate',
    '\S\+\ze invalid because',
    'contained',
    containedin='MyoSplain',
)
splain_foundreq = SyntaxMatch.cons(
    'MyoSplainFoundReq',
    f'^.*{found_marker}.\{{-}}{separator_marker}{foundreq_separator}.\{{-}}{req_marker}.*$',
    'contained',
    contains='MyoSplainFound,MyoSplainReq',
)
splain_found = SyntaxMatch.cons(
    'MyoSplainFound',
    f'{found_marker}.\{{-}}{separator_marker}',
    'contained',
    contains='MyoSplainFoundReqMarker',
    nextgroup='MyoSplainReq',
)
splain_req = SyntaxMatch.cons(
    'MyoSplainReq',
    f'{foundreq_separator}\zs.\{{-}}\ze{req_marker}',
    'contained',
    nextgroup='MyoSplainFoundReqMarker',
)
splain_req_marker_conceal = conceal('MyoSplainFoundReqMarker', f'{found_marker}\|{separator_marker}\|{req_marker}')

hl_error = HiLink('MyoError', 'Error')
hl_path = HiLink('MyoPath', 'Directory')
hl_line_number = HiLink('MyoLineNumber', 'Directory')
hl_col = HiLink('MyoCol', 'Search')
hl_splain_param_marker_bang = HiLink('MyoSplainParamMarkerBang', 'Error')
hl_splain_param_marker_i = HiLink('MyoSplainParamMarkerI', 'Directory')
hl_splain_param_name = HiLink('MyoSplainParamName', 'Type')
hl_splain_param_type = HiLink('MyoSplainParamType', 'Statement')
hl_splain_candidate = HiLink('MyoSplainCandidate', 'Error')
hl_splain_found_req = HiLink('MyoSplainFoundReq', 'Normal')
hl_splain_found = HiLink('MyoSplainFound', 'Error')

hi_req = Highlight.cons('MyoSplainReq', ctermfg=2, guifg='#719e07')

syntax = Syntax(
    List(scala_include, location, path, line_number, error, code_line, col_marker_conceal, col, splain, splain_param,
         splain_param_marker, splain_param_marker_bang, splain_param_marker_i, splain_param_name, splain_param_type,
         splain_candidate, splain_foundreq, splain_found, splain_req, splain_req_marker_conceal),
    List(hi_req),
    List(hl_error, hl_path, hl_line_number, hl_col, hl_splain_param_marker_bang, hl_splain_param_marker_i,
         hl_splain_param_name, hl_splain_param_type, hl_splain_candidate, hl_splain_found),
)


def scala_syntax() -> NS[CommandRibosome, Syntax]:
    return NS.pure(syntax)


__all__ = ('scala_syntax',)



from chiasma.util.id import IdentSpec, ensure_ident_or_generate

from amino import do, Do
from amino.logging import module_log
from amino.state import EitherState

from ribosome.compute.api import prog
from ribosome.compute.ribosome_api import Ribo

from myo.util import Ident
from myo.ui.data.ui_data import UiData
from myo.ui.data.window import Window
from myo.ui.pane import open_or_toggle_pane, map_window_trees
from myo.components.ui.compute.pane import render_view

log = module_log()


@do(EitherState[str, UiData, Window])
def ui_toggle_pane(ident: Ident) -> Do:
    yield map_window_trees(lambda a: open_or_toggle_pane(ident)(a).get_or_strict(a)).nvim


@prog.do(None)
def toggle_pane(ident_spec: IdentSpec) -> Do:
    ident = ensure_ident_or_generate(ident_spec)
    yield Ribo.lift_comp(ui_toggle_pane(ident), UiData)
    yield render_view(ident)


__all__ = ('toggle_pane',)



from ribosome.compute.ribosome import Ribosome

from myo.env import Env
from myo.config.component import MyoComponent
from myo.ui.data.ui_data import UiData

UiRibosome = Ribosome[Env, MyoComponent, UiData]


__all__ = ('UiRibosome',)



from amino import ADT, Path, Dat

from myo.data.command import Command
from myo.ui.data.view import Pane


class RunTaskDetails(ADT['RunTaskDetails']):
    pass


class VimTaskDetails(RunTaskDetails):
    pass


class SystemTaskDetails(RunTaskDetails):
    pass


class UiSystemTaskDetails(RunTaskDetails):

    def __init__(self, pane: Pane) -> None:
        self.pane = pane


class UiShellTaskDetails(RunTaskDetails):

    def __init__(self, shell: Command, pane: Pane) -> None:
        self.shell = shell
        self.pane = pane


class RunTask(Dat['RunTask']):

    def __init__(self, command: Command, log: Path, details: RunTaskDetails) -> None:
        self.command = command
        self.log = log
        self.details = details


def is_system_task(task: RunTaskDetails) -> bool:
    return isinstance(task, (SystemTaskDetails, UiSystemTaskDetails))


__all__ = ('RunTaskDetails', 'VimTaskDetails', 'SystemTaskDetails', 'UiSystemTaskDetails', 'UiShellTaskDetails',
           'RunTask', 'is_system_task',)



from typing import Union

from networkx import DiGraph

from amino import List, Path, do, Either, Try, Do, Maybe, Regex, Nil, ADT, Just, Dat
from amino.util.numeric import parse_int
from amino.logging import module_log

from myo.output.data.output import OutputLine, OutputEvent, Location
from myo.output.parser.base import EdgeData, Parser

log = module_log()


class HaskellLine(ADT['HaskellLine']):
    pass


class FileLine(HaskellLine):

    @staticmethod
    @do(Either[str, 'FileLine'])
    def cons(
            path: Union[str, Path],
            line: Union[str, int],
            col: Union[str, int]=None,
    ) -> Do:
        path_p = yield Try(Path, path)
        line_i = yield parse_int(line)
        col_i = yield parse_int(1 if col is None else col)
        return FileLine(path_p, line_i - 1, col_i - 1)

    def __init__(self, path: Path, line: int, col: int) -> None:
        self.path = path
        self.line = line
        self.col = col


class InfoLine(HaskellLine):

    @staticmethod
    def cons(message: str, ws: str='', tag: str='') -> 'InfoLine':
        return InfoLine(message, len(ws), tag)

    def __init__(self, message: str, indent: int, tag: str) -> None:
        self.message = message
        self.indent = indent
        self.tag = tag


class CodeLine(HaskellLine):

    def __init__(self, code: str, indent: int, tag: str) -> None:
        self.code = code
        self.indent = indent
        self.tag = tag


_msg_type = '\[(?P<tag>\w+)\] '
file_edge: EdgeData[FileLine] = EdgeData(
    regex=Regex(f'^(?P<path>/[^:]+):(?P<line>\d+):((?P<col>\d+):)?'),
    cons_output_line=FileLine.cons,
)
info_edge = EdgeData.strict(
    regex=Regex(f'^(?P<ws>\s*)(?P<message>\s*.+)'),
    cons_output_line=InfoLine.cons,
)


class HaskellEvent(Dat['HaskellEvent']):

    @staticmethod
    def cons(
            file: OutputLine[FileLine],
            info: List[OutputLine[InfoLine]]=Nil,
    ) -> 'HaskellEvent':
        return HaskellEvent(file, info)

    def __init__(
            self,
            file: OutputLine[FileLine],
            info: List[OutputLine[InfoLine]],
    ) -> None:
        self.file = file
        self.info = info


def haskell_graph() -> DiGraph:
    g = DiGraph()  # type: ignore
    g.add_edge('start', 'file', data=file_edge)
    g.add_edge('file', 'info', data=info_edge)
    g.add_edge('info', 'info', data=info_edge)
    return g


@do(Maybe[OutputEvent[HaskellLine, HaskellEvent]])
def haskell_event(lines: List[OutputLine[HaskellLine]]) -> Do:
    file = yield lines.find(lambda a: isinstance(a.meta, FileLine))
    info = lines.filter(lambda a: isinstance(a.meta, InfoLine))
    location = Location(file.meta.path, file.meta.line, file.meta.col)
    return OutputEvent.cons(HaskellEvent(file, info), lines, Just(location))


def haskell_events(lines: List[OutputLine[HaskellLine]]) -> List[OutputEvent[HaskellLine, HaskellEvent]]:
    return haskell_event(lines).cata_strict(List, Nil)


haskell_parser = Parser(haskell_graph(), haskell_events)

__all__ = ('haskell_parser',)



from typing import Tuple

from amino import do, Do, Just, Nothing, List, Nil

from chiasma.util.id import IdentSpec
from chiasma.data.view_tree import find_in_view_tree, ViewTree, SubUiNode
from amino.lenses.lens import lens
from amino.logging import module_log
from amino.state import EitherState
from amino.case import Case

from ribosome.compute.api import prog
from ribosome.compute.ribosome_api import Ribo

from myo.util import Ident
from myo.ui.data.ui_data import UiData
from myo.ui.data.window import Window
from myo.components.ui.compute.pane import render_view
from myo.ui.data.view import Layout, Pane
from myo.ui.pane import ui_modify_layout_node
from myo.ui.data.view_tree import MyoLayoutNode, MyoViewTree, MyoPaneNode
from myo.util.id import ensure_ident_prog

log = module_log()


class open_first_pane(Case[MyoViewTree, Tuple[MyoViewTree, bool]], alg=ViewTree):

    def layout(self, node: MyoLayoutNode) -> Tuple[MyoViewTree, bool]:
        def folder(z: Tuple[List[MyoViewTree], bool], a: MyoViewTree) -> Tuple[MyoViewTree, bool]:
            old, found_pre = z
            new, found_post = (
                (a, True)
                if found_pre else
                self(a)
            )
            return old.cat(new), found_post
        new, found = node.sub.fold_left((Nil, False))(folder)
        v = node.set.sub(new)
        return v, found

    def pane(self, a: MyoPaneNode) -> Tuple[MyoViewTree, bool]:
        return lens.data.open.set(True)(a), True

    def sub_ui(self, a: SubUiNode[Layout, Pane]) -> Tuple[MyoViewTree, bool]:
        return a, False


def layout_open(node: MyoLayoutNode) -> MyoLayoutNode:
    return find_in_view_tree(pane=lambda a: Just(True) if a.data.open else Nothing)(node).present


def open_layout(node: MyoLayoutNode) -> MyoLayoutNode:
    new, found = open_first_pane.match(node)
    return new


def open_or_toggle_minimized(node: MyoLayoutNode) -> MyoLayoutNode:
    return (
        lens.data.state.minimized.modify(lambda a: ~a)(node)
        if layout_open(node) else
        open_layout(node)
    )


@do(EitherState[str, UiData, Window])
def ui_toggle_layout(ident: Ident) -> Do:
    yield ui_modify_layout_node(ident, open_or_toggle_minimized)
    # yield map_window_trees(lambda a: open_or_toggle_layout(ident)(a).get_or_strict(a)).nvim


@prog.do(None)
def toggle_layout(ident_spec: IdentSpec) -> Do:
    ident = yield ensure_ident_prog(ident_spec)
    yield Ribo.lift_comp(ui_toggle_layout(ident).nvim, UiData)
    yield render_view(ident)


__all__ = ('toggle_layout',)



from amino import do, Do, Maybe, List
from amino.logging import module_log
from amino.lenses.lens import lens

from chiasma.util.id import Ident

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.util.persist import store_json_state, load_json_state
from ribosome.compute.ribosome_api import Ribo

from myo.data.command import Command, HistoryEntry
from myo.components.command.data import CommandData
from myo.components.command.compute.tpe import CommandRibosome

log = module_log()


def store_history() -> NS[CommandData, None]:
    return store_json_state('history', lambda a: a.history)


def restore_history() -> NS[CommandData, None]:
    return load_json_state('history', lens.history)


def insert_history_entry(history: List[HistoryEntry], entry: HistoryEntry) -> List[HistoryEntry]:
    return history.cons(entry).distinct_by(lambda a: (a.cmd.ident, a.cmd.lines))


@prog
@do(NS[CommandRibosome, None])
def push_history(cmd: Command, target: Maybe[Ident]) -> Do:
    entry = HistoryEntry(cmd, target)
    yield Ribo.modify_comp(lens.history.modify(lambda a: insert_history_entry(a, entry)))
    yield Ribo.zoom_comp(store_history())


@prog
def history() -> NS[CommandRibosome, List[HistoryEntry]]:
    return Ribo.inspect_comp(lambda a: a.history)


__all__ = ('push_history', 'history', 'store_history', 'restore_history',)
